"""
3 traces vertically covering the regions:
- J: 1.0, 1.35     dwave = 0.35 --> R~2800
- H: 1.3, 1.8      dwave = 0.5  --> R~2000
- K: 1.75, 2.5     dwave = 0.75 --> R~1300

The detector is 1024 * 0.2 = 204.8" [+/-102.4"] and 10.24mm [+/-5.12mm]
Slit is 60", or 300 pixel, so trace x-borders should be:
- J: (-95, -35) ["], --> /20 "/mm --> (-4.75, -1.75) [mm] centred at -3.25
- H: (-30,  30) ["], --> /20 "/mm --> (-1.5,   1.5)  [mm] centred at 0
- K: ( 35,  95) ["], --> /20 "/mm --> ( 1.75,  4.75) [mm] centred at 3.25

Oliver's SpectralTraces follow a O(5) polynomial, therefore we need 5
points along the slits
s = [-30, -10, 0, 10, 30] ["]
dx = [-1.5, -0.5, 0, 0.5, 1.5 ] [mm]

The dispersion should follow a 10-point logspace increase from
wave_min, wave_max for each of the trace extremes

y value go *linspace* from +/-500 px --> +/-5 mm

The format of the fits file is:
wave, slit ["], x [mm], y [mm]

"""
from copy import deepcopy
import numpy as np
from astropy.table import Table
from astropy import units as u
from astropy.io import fits
from matplotlib import pyplot as plt


def make_lss_trace_file():
    names = ["J", "H", "K"]
    wave_mins = np.array([1.0, 1.3, 1.75])
    wave_maxs = np.array([1.35, 1.8, 2.5])
    s_cens = np.array([-65., 0., 65.])
    dss = np.array([-30., 10., 0., 10., 30.])

    hdus = []
    for wave_min, wave_max, s_cen, name in zip(wave_mins, wave_maxs, s_cens, names):
        n = 101

        waves = np.geomspace(wave_min, wave_max, n)
        waves = np.array([[w]*5 for w in waves]).flatten()

        ys = np.array([[y]*5 for y in np.linspace(-5, 5, n)]).flatten()

        ss = np.array(list(dss) * n)
        xs = (ss + s_cen) / 20.

        tbl = Table(names=["wavelength", "s", "x", "y"],
                    data=[waves, ss, xs, ys],
                    units=[u.um, u.arcsec, u.mm, u.mm])
        hdu = fits.table_to_hdu(tbl)
        hdu.header["EXTNAME"] = f"TRACE_{name}"
        hdus += [hdu]

    tbl = Table(names=["description", "extension_id", "aperture_id", "image_plane_id"],
                data=[[hdu.header["EXTNAME"] for hdu in hdus],
                      [2,3,4],
                      [0,0,0],
                      [0,0,0]])
    cat_hdu = fits.table_to_hdu(tbl)
    cat_hdu.header["EXTNAME"] = "TOC"

    pri_hdr = fits.PrimaryHDU()
    pri_hdr.header["ECAT"] = 1         # where to find the catalogue table
    pri_hdr.header["EDATA"] = 2        # where the data tables start

    hdul = fits.HDUList([pri_hdr, cat_hdu] + hdus)
    hdul.writeto("../INS_lss_traces.fits", overwrite=True)

# make_lss_trace_file()

def make_ifu_trace_file():
    names = [f"Ap{i}" for i in range(5)]
    wave_min = 1.75
    wave_max = 2.5
    s_cens = np.array([-64., -32., 0., 32., 64.])   # across detector
    dss = np.array([-14., -7., 0., 7., 14.])     # across slit

    hdus = []
    for s_cen, name in zip(s_cens, names):
        n = 101

        waves = np.geomspace(wave_min, wave_max, n)
        waves = np.array([[w] * 5 for w in waves]).flatten()    # 5 pos along slit

        ys = np.array([[y] * 5 for y in np.linspace(-5, 5, n)]).flatten()

        ss = np.array(list(dss) * n)
        xs = (ss + s_cen) / 20.

        tbl = Table(names=["wavelength", "s", "x", "y"],
                    data=[waves, ss, xs, ys],
                    units=[u.um, u.arcsec, u.mm, u.mm])
        hdu = fits.table_to_hdu(tbl)
        hdu.header["EXTNAME"] = f"TRACE_{name}"
        hdus += [hdu]

    tbl = Table(
        names=["description", "extension_id", "aperture_id", "image_plane_id"],
        data=[[hdu.header["EXTNAME"] for hdu in hdus],
              [2, 3, 4, 5, 6],
              [0, 1, 2, 3, 4],
              [0, 0, 0, 0, 0]])
    cat_hdu = fits.table_to_hdu(tbl)
    cat_hdu.header["EXTNAME"] = "TOC"

    pri_hdr = fits.PrimaryHDU()
    pri_hdr.header["ECAT"] = 1  # where to find the catalogue table
    pri_hdr.header["EDATA"] = 2  # where the data tables start

    hdul = fits.HDUList([pri_hdr, cat_hdu] + hdus)
    hdul.writeto("../INS_ifu_traces.fits", overwrite=True)

# make_ifu_trace_file()



import os
from astropy.table import Table

from ..effects.data_container import DataContainer
from .. import base_classes as bc
from ..utils import from_currsys, write_report
from ..reports.rst_utils import table_to_rst
from .. import rc


class Effect(DataContainer):
    """
    The base class for representing the effects (artifacts) in an optical system

    The ``Effect`` class is conceived to independently apply the changes that
    an optical component (or series thereof) has on an incoming 3D description
    of an on-sky object. In other words, **an Effect object should receive a
    derivative of a ``Source`` object, alter it somehow, and return it**.

    The interface for the Effect base-class has been kept very general so that
    it can easily be sub-classed as data for new effects becomes available.
    Essentially, a sub-classed Effects object must only contain the following
    attributes:

    * ``self.meta`` - a dictionary to contain meta data.
    * ``self.apply_to(obj, **kwargs)`` - a method which accepts a
      Source-derivative and returns an instance of the same class as ``obj``
    * ``self.fov_grid(which="", **kwargs)``


    Parameters
    ----------
    See :class:`DataContainer` for input parameters

    Methods


    """

    def __init__(self, **kwargs):
        super(Effect, self).__init__(**kwargs)
        self.meta["z_order"] = []
        self.meta["include"] = True
        self.meta.update(kwargs)

    def apply_to(self, obj, **kwargs):
        if not isinstance(obj, (bc.FOVSetupBase, bc.SourceBase,
                                bc.FieldOfViewBase, bc.ImagePlaneBase,
                                bc.DetectorBase)):
            raise ValueError(f"object must one of the following: FOVSetupBase, "
                             f"Source, FieldOfView, ImagePlane, Detector: "
                             f"{type(obj)}")

        return obj

    def fov_grid(self, which="", **kwargs):
        """
        Returns the edges needed to generate FieldOfViews for an observation

        Parameters
        ----------
        which : str
            ["waveset", "edges", "shifts"] where:
            * waveset - wavelength bin extremes
            * edges - on sky coordinate edges for each FOV box
            * shifts - wavelength dependent FOV position offsets

        kwargs
        ------
        wave_min, wave_max : float
            [um] list of wavelength

        wave_mid : float
            [um] wavelength what will be centred on optical axis


        Returns
        -------
        waveset : list
            [um] N+1 wavelengths that set edges of N spectral bins

        edges : list of lists
            [arcsec] Contains a list of footprint lists

        shifts : list of 3 lists
            [wave, dx, dy] Contains lists corresponding to the (dx, dy) offset
            from the optical axis (0, 0) induced for each wavelength in (wave)
            [um, arcsec, arcsec]

        """
        self.update(**kwargs)
        return []

    def update(self, **kwargs):
        self.meta.update(kwargs)
        # self.update_bang_keywords()

    # def update_bang_keywords(self):
    #     for key in self.meta:
    #         if isinstance(self.meta[key], str) and self.meta[key][0] == "!":
    #             bang_key = self.meta[key]
    #             self.meta[key] = rc.__currsys__[bang_key]

    @property
    def include(self):
        return from_currsys(self.meta["include"])

    @include.setter
    def include(self, item):
        self.meta["include"] = item

    @property
    def display_name(self):
        return self.meta.get("name", self.meta.get("filename", "<empty>"))

    @property
    def meta_string(self):
        meta_str = ""
        max_key_len = max([len(key) for key in self.meta.keys()])
        for key in self.meta:
            if key not in ["comments", "changes", "description", "history",
                           "report_table_caption", "report_plot_caption",
                           "table"]:
                meta_str += "    {} : {}\n".format(key.rjust(max_key_len),
                                                   self.meta[key])

        return meta_str

    def report(self, filename=None, output="rst", rst_title_chars="*+",
               **kwargs):
        """
        For Effect objects, generates a report based on the data and meta-data

        This is to aid in the automation of the documentation process of the
        instrument packages in the IRDB.

        .. note:: If the Effect can generate a plot, this will be saved to disc

        Parameters
        ----------
        filename : str, optional
            Where to save the RST file
        output : str, optional
            ["rst", "latex"] Output file format
        rst_title_chars : 2-str, optional
            Two unique characters used to denote rst subsection headings.
            Options: = - ` : ' " ~ ^ _ * + # < >

        Additional parameters
        ---------------------
        Either from the ``self.meta["report"]`` dictionary or via ``**kwargs``

        "report_table_include": False
        "report_table_caption": ""
        "report_plot_caption": ""
        "report_plot_include": False
        "report_plot_file_formats": ["png"]
            Multiple formats can be saved. The last entry is used for the RST.
        "report_plot_filename": None
            If None, uses self.meta["name"] as the filename
        "file_description": str
            Taken from the header of a file, if available
        "class_description": str
            Taken from the docstring of the subclass
        "changes_str": list of str
            Take from the header of a file, if available

        Returns
        -------
        rst_str : str
            The full reStructureText string

        Notes
        -----

        The format of the RST output is as follows::

            <ClassType>: <effect name>
            **************************
            File Description: <description for file meta data>
            Class Description: <description from class docstring>
            Changes: <list of changes from file meta data>

            Data
            ++++
            .. figure:: <Figure_name>.png
                If the <Effect> object contains a ``.plot()`` function, add
                plot and write it to disc
            Figure caption

            Table caption
            Table
                If the <Effect> object contains a ``.table()`` function, add
                a pprint version of the table

            Meta-data
            +++++++++
            ::
                A code block print out of the ``.meta`` dictionary


        """
        changes = self.meta.get("changes", [])
        changes_str = "- " + "\n- ".join([str(entry) for entry in changes])
        cls_doc = self.__doc__ if self.__doc__ is not None else "<no docstring>"
        cls_descr = cls_doc.lstrip().splitlines()[0]

        params = {"report_plot_filename": None,
                  "report_plot_file_formats": ["png"],
                  "report_plot_caption": "",
                  "report_plot_include": False,
                  "report_table_include": False,
                  "report_table_caption": "",
                  "report_table_rounding": None,
                  "report_image_path": "!SIM.reports.image_path",
                  "report_rst_path": "!SIM.reports.rst_path",
                  "report_latex_path": "!SIM.reports.latex_path",
                  "file_description": self.meta.get("description",
                                                    "<no description>"),
                  "class_description": cls_descr,
                  "changes_str": changes_str}
        params.update(self.meta)
        params.update(kwargs)
        params = from_currsys(params)

        rst_str = """
{}
{}
**Included by default**: ``{}``

**File Description**: {}

**Class Description**: {}

**Changes**:

{}

Data
{}
""".format(str(self),
           rst_title_chars[0] * len(str(self)),
           params["include"],
           params["file_description"],
           params["class_description"],
           params["changes_str"],
           rst_title_chars[1] * 4)

        if params["report_plot_include"] and hasattr(self, "plot"):
            fig = self.plot()

            if fig is not None:
                path = params["report_image_path"]
                fname = params["report_plot_filename"]
                if fname is None:
                    fname = self.meta["name"].lower().replace(" ", "_")

                for fmt in params["report_plot_file_formats"]:
                    fname = ".".join((fname.split(".")[0], fmt))
                    file_path = os.path.join(path, fname)

                    fig.savefig(fname=file_path)

                # rel_path = os.path.relpath(params["report_image_path"],
                #                            params["report_rst_path"])
                # rel_file_path = os.path.join(rel_path, fname)

                rst_str += """
.. figure:: {}
    :name: {}

    {}
""".format(fname,
           "fig:" + params.get("name", "<unknown Effect>"),
           params["report_plot_caption"])

        if params["report_table_include"]:
            rst_str += """
.. table::
    :name: {}

{}

{}
""".format("tbl:" + params.get("name"),
           table_to_rst(self.table, indent=4,
                        rounding=params["report_table_rounding"]),
           params["report_table_caption"])

        rst_str += """
Meta-data
{}
::

{}
""".format(rst_title_chars[1] * 9,
           self.meta_string)

        write_report(rst_str, filename, output)

        return rst_str

    def info(self):
        """
        Prints basic information on the effect, notably the description
        """
        name = self.meta.get("name", self.meta.get("filename", "<empty>"))
        text = f'{type(self).__name__}: "{name}"'

        desc = self.meta.get("description")
        if desc is not None:
             text += f"\nDescription: {desc}"

        print(text)

    def __repr__(self):
        return f'{type(self).__name__}: "{self.display_name}"'

    def __str__(self):
        return self.__repr__()

    def __getitem__(self, item):
        if isinstance(item, str) and item[0] == "#":
            if len(item) > 1:
                if item[-1] == "!":
                    key = item[1:-1]
                    if len(key) > 0:
                        value = from_currsys(self.meta[key])
                    else:
                        value = from_currsys(self.meta)
                else:
                    value = self.meta[item[1:]]
            else:
                value = self.meta
        else:
            raise ValueError(f"__getitem__ calls must start with '#': {item}")

        return value


"""
Functions to download instrument packages and example data
"""
import json
import re
import shutil
import os
import urllib.request
import zipfile
import logging
from urllib3.exceptions import HTTPError

import yaml
import requests
import bs4
from astropy.utils.data import download_file

from scopesim import rc


def get_server_package_list():
    url = rc.__config__["!SIM.file.server_base_url"]
    response = requests.get(url + "packages.yaml")
    pkgs_dict = yaml.full_load(response.text)

    return pkgs_dict


def get_server_folder_contents(dir_name, unique_str=".zip"):
    url = rc.__config__["!SIM.file.server_base_url"] + dir_name

    try:
        result = requests.get(url).content
    except Exception as error:
        raise ValueError(f"URL returned error: {url}") from error

    soup = bs4.BeautifulSoup(result, features="lxml")
    hrefs = soup.findAll("a", href=True)
    pkgs = [href.string for href in hrefs
            if href.string is not None and ".zip" in href.string]

    return pkgs


def list_packages(pkg_name=None):
    """
    List all packages, or all variants of a single package

    Parameters
    ----------
    pkg_name : str, optional
        - None: lists all stable packages on the server
        - <PackageName>: lists all variants of <PackageName> on the server

    Returns
    -------
    pkg_names : list

    Examples
    --------
    ::
        from scopesim import list_packages

        # list all stable packages on the server
        list_packages()

        # list all variants of a specific package
        list_packages("Armazones")

    """
    pkgs_dict = get_server_package_list()

    if pkg_name is None:
        pkg_names = list(pkgs_dict.keys())
    elif pkg_name in pkgs_dict:
        path = pkgs_dict[pkg_name]["path"]
        pkgs = get_server_folder_contents(path)
        pkg_names = [pkg for pkg in pkgs if pkg_name in pkg]

    return pkg_names


def download_packages(pkg_names, release="stable", save_dir=None, from_cache=None):
    """
    Download one or more packages to the local disk

    1. Download stable, dev
    2. Download specific version
    3. Download from github via url

    Parameters
    ----------
    pkg_names : str, list
        A list of package names, see ``list_packages()``

    release : str, optional
        By default, the most recent stable version of a package is downloaded.
        Other options are:
        - "stable" : the most recent stable version
        - "latest" : the latest development version to be published
        - a specific package filename as given by list_packages (see examples)
        - a github url for the specific branch and package (see examples)

    save_dir : str, optional
        The place on the local disk where the ``.zip`` package is to be saved.
        If left as None, defaults to the value in
        scopesim.rc.__config__["!SIM.file.local_packages_path"]

    from_cache : bool, optional
        Use the cached versions of the packages. If None, defaults to the RC
        value: ``!SIM.file.use_cached_downloads``

    Returns
    -------
    save_path : str
        The absolute path to the saved ``.zip`` package

    Examples
    --------
    ::
        from scopesim import download_packages, list_packages

        # Stable release of a list of packages
        download_packages(["test_package", "test_package"])

        # Development release of a single package
        download_packages("test_package", release="latest")

        # Specific version of the package
        list_packages("test_package")
        download_packages("test_package", release="2022-04-09.dev")

        # Specific package from a Gtihub commit hash or branch/tag name (use "@" or ":")
        download_packages("ELT", release="github:728761fc76adb548696205139e4e9a4260401dfc")
        download_packages("ELT", release="github@728761fc76adb548696205139e4e9a4260401dfc")
        download_packages("ELT", release="github@dev_master")

    """
    base_url = rc.__config__["!SIM.file.server_base_url"]

    pkgs_dict = get_server_package_list()

    if isinstance(pkg_names, str):
        pkg_names = [pkg_names]

    save_paths = []
    for pkg_name in pkg_names:
        if pkg_name in pkgs_dict:
            pkg_dict = pkgs_dict[pkg_name]
            path = pkg_dict["path"] + "/"

            from_github = False
            if release in ["stable", "latest"]:
                zip_name = pkg_dict[release]
                pkg_url = f"{base_url}{path}/{zip_name}.zip"
            elif "github" in release:
                base_url = "https://github.com/AstarVienna/irdb/tree/"
                github_hash = release.split(":")[-1].split("@")[-1]
                pkg_url = f"{base_url}{github_hash}/{pkg_name}"
                from_github = True
            else:
                zip_name = f"{pkg_name}.{release}.zip"
                pkg_variants = get_server_folder_contents(path)
                if zip_name not in pkg_variants:
                    raise ValueError(f"{zip_name} is not amoung the hosted "
                                     f"variants: {pkg_variants}")
                pkg_url = f"{base_url}{path}/{zip_name}"

            if save_dir is None:
                save_dir = rc.__config__["!SIM.file.local_packages_path"]
            if not os.path.exists(save_dir):
                os.mkdir(save_dir)

            if not from_github:
                try:
                    if from_cache is None:
                        from_cache = rc.__config__["!SIM.file.use_cached_downloads"]
                    cache_path = download_file(pkg_url, cache=from_cache)
                    save_path = os.path.join(save_dir, f"{pkg_name}.zip")
                    file_path = shutil.copy2(cache_path, save_path)

                    with zipfile.ZipFile(file_path, 'r') as zip_ref:
                        zip_ref.extractall(save_dir)

                except HTTPError as error:
                    raise ValueError(f"Unable to find file: {url + pkg_path}") from error
            else:
                download_github_folder(repo_url=pkg_url, output_dir=save_dir)
                save_path = save_dir

            save_paths += [os.path.abspath(save_path)]

        else:
            raise HTTPError(f"Unable to find package: {base_url + pkg_name}")

    return save_paths

# ==============================================================================
# Funtions below from from OLD_database.py
# ==============================================================================

# for backwards compatibility
def download_package(pkg_path, save_dir=None, url=None, from_cache=None):
    """
    Downloads a package to the local disk

    Parameters
    ----------
    pkg_path : str, list
        A ``.zip`` package path as given by ``list_packages()``

    save_dir : str
        The place on the local disk where the ``.zip`` package is to be saved.
        If left as None, defaults to the value in
        scopesim.rc.__config__["!SIM.file.local_packages_path"]

    url : str
        The URL of the IRDB HTTP server. If left as None, defaults to the
        value in scopesim.rc.__config__["!SIM.file.server_base_url"]

    from_cache : bool
        Use the cached versions of the packages. If None, defaults to the RC
        value: ``!SIM.file.use_cached_downloads``

    Returns
    -------
    save_path : str
        The absolute path to the saved ``.zip`` package

    """
    # todo: add proper depreciation warning
    text = "Function Depreciated --> please use scopesim.download_package-s-()"
    logging.warning(text)
    print(text)

    if isinstance(pkg_path, str):
        pkg_path = [pkg_path]

    pkg_names = [pkg.replace(".zip", "").split("/")[-1] for pkg in pkg_path]
    return download_packages(pkg_names, release="stable", save_dir=save_dir,
                             from_cache=from_cache)

def get_server_elements(url, unique_str="/"):
    """
    Returns a list of file and/or directory paths on the HTTP server ``url``

    Parameters
    ----------
    url : str
        The URL of the IRDB HTTP server.

    unique_str : str, list
        A unique string to look for in the beautiful HTML soup:
        "/" for directories this, ".zip" for packages

    Returns
    -------
    paths : list
        List of paths containing in ``url`` which contain ``unique_str``

    """
    if isinstance(unique_str, str):
        unique_str = [unique_str]

    try:
        result = requests.get(url).content
    except Exception as error:
        raise ValueError(f"URL returned error: {url}") from error

    soup = bs4.BeautifulSoup(result, features="lxml")
    paths = soup.findAll("a", href=True)
    select_paths = []
    for the_str in unique_str:
        select_paths += [tmp.string for tmp in paths
                         if tmp.string is not None and the_str in tmp.string]
    return select_paths


def list_example_data(url=None, return_files=False, silent=False):
    """
    List all example files found under ``url``

    Parameters
    ----------
    url : str
        The URL of the database HTTP server. If left as None, defaults to the
        value in scopesim.rc.__config__["!SIM.file.server_base_url"]

    return_files : bool
        If True, returns a list of file names

    silent : bool
        If True, does not print the list of file names

    Returns
    -------
    all_files : list of str
        A list of paths to the example files relative to ``url``.
        The full string should be passed to ``download_example_data``.
    """

    def print_file_list(the_files, loc=""):
        print(f"\nFiles saved {loc}\n" + "=" * (len(loc) + 12))
        for _file in the_files:
            print(_file)

    if url is None:
        url = rc.__config__["!SIM.file.server_base_url"]

    return_file_list = []
    server_files = []
    folders = get_server_elements(url, "example_data")
    for folder in folders:
        files = get_server_elements(url + folder, ("fits", "txt", "dat"))
        server_files += files
    if not silent:
        print_file_list(server_files, f"on the server: {url + 'example_data/'}")
        return_file_list += server_files

    if return_files:
        return return_file_list

    return None


def download_example_data(file_path, save_dir=None, url=None, from_cache=None):
    """
    Downloads example fits files to the local disk

    Parameters
    ----------
    file_path : str, list
        Name(s) of FITS file(s) as given by ``list_example_data()``

    save_dir : str
        The place on the local disk where the downloaded files are to be saved.
        If left as None, defaults to the current working directory.

    url : str
        The URL of the database HTTP server. If left as None, defaults to the
        value in scopesim.rc.__config__["!SIM.file.server_base_url"]

    from_cache : bool
        Use the cached versions of the files. If None, defaults to the RC
        value: ``!SIM.file.use_cached_downloads``

    Returns
    -------
    save_path : str
        The absolute path to the saved files
    """
    if isinstance(file_path, (list, tuple)):
        save_path = [download_example_data(thefile, save_dir, url)
                     for thefile in file_path]
    elif isinstance(file_path, str):

        if url is None:
            url = rc.__config__["!SIM.file.server_base_url"]
        if save_dir is None:
            save_dir = os.getcwd()
        if not os.path.exists(save_dir):
            os.mkdir(save_dir)

        try:
            if from_cache is None:
                from_cache = rc.__config__["!SIM.file.use_cached_downloads"]
            cache_path = download_file(url + "example_data/" + file_path,
                                       cache=from_cache)
            save_path = os.path.join(save_dir, os.path.basename(file_path))
            file_path = shutil.copy2(cache_path, save_path)
        except HTTPError:
            ValueError(f"Unable to find file: {url + 'example_data/' + file_path}")

        save_path = os.path.abspath(save_path)

    return save_path


# """
# 2022-04-10 (KL)
# Code taken directly from https://github.com/sdushantha/gitdir
# Adapted for ScopeSim usage.
# Many thanks to the authors!
# """

def create_github_url(url):
    """
    From the given url, produce a URL that is compatible with Github's REST API. Can handle blob or tree paths.
    """
    repo_only_url = re.compile(r"https:\/\/github\.com\/[a-z\d](?:[a-z\d]|-(?=[a-z\d])){0,38}\/[a-zA-Z0-9]+$")
    re_branch = re.compile("/(tree|blob)/(.+?)/")

    # Check if the given url is a url to a GitHub repo. If it is, tell the
    # user to use 'git clone' to download it
    if re.match(repo_only_url,url):
        message = "✘ The given url is a complete repository. Use 'git clone' to download the repository"
        logging.error(message)
        raise ValueError(message)

    # extract the branch name from the given url (e.g master)
    branch = re_branch.search(url)
    download_dirs = url[branch.end():]
    api_url = (url[:branch.start()].replace("github.com", "api.github.com/repos", 1) +
              "/contents/" + download_dirs + "?ref=" + branch.group(2))
    return api_url, download_dirs


def download_github_folder(repo_url, output_dir="./"):
    """
    Downloads the files and directories in repo_url.

    Re-written based on the on the download function `here <https://github.com/sdushantha/gitdir/blob/f47ce9d85ee29f8612ce5ae804560a12b803ddf3/gitdir/gitdir.py#L55>`_
    """
    # convert repo_url into an api_url
    api_url, download_dirs = create_github_url(repo_url)

    # get the contents of the github folder
    user_interrupt_text = "GitHub download interrupted by User"
    try:
        opener = urllib.request.build_opener()
        opener.addheaders = [('User-agent', 'Mozilla/5.0')]
        urllib.request.install_opener(opener)
        response = urllib.request.urlretrieve(api_url)
    except KeyboardInterrupt:
        # when CTRL+C is pressed during the execution of this script
        logging.error(user_interrupt_text)
        raise ValueError(user_interrupt_text)

    # Make the base directories for this GitHub folder
    os.makedirs(os.path.join(output_dir, download_dirs), exist_ok=True)

    with open(response[0], "r") as f:
        data = json.load(f)

        for entry in data:
            # if the entry is a further folder, walk through it
            if entry["type"] == "dir":
                download_github_folder(repo_url=entry["html_url"],
                                       output_dir=output_dir)

            # if the entry is a file, download it
            elif entry["type"] == "file":
                try:
                    opener = urllib.request.build_opener()
                    opener.addheaders = [('User-agent', 'Mozilla/5.0')]
                    urllib.request.install_opener(opener)
                    # download the file
                    save_path = os.path.join(output_dir, entry['path'])
                    urllib.request.urlretrieve(entry["download_url"], save_path)
                    logging.info(f"Downloaded: {entry['path']}")

                except KeyboardInterrupt:
                    # when CTRL+C is pressed during the execution of this script
                    logging.error(user_interrupt_text)
                    raise ValueError(user_interrupt_text)



import numpy as np

from ..base_classes import PoorMansHeader


class MonochromeTraceCurve:
    """
    Contains coordinates along a monochromatic section of a spectral trace

    Used to generate detector plane position WCS info for FOVs in spectral mode
    """
    def __init__(self, x, y, s, wave_min, wave_max, **kwargs):
        self.wave_min = wave_min
        self.wave_max = wave_max
        self.x = x
        self.y = y
        self.s = s

        self.meta = {"rotation": 0,
                     "shear": 0}
        self.meta.update(kwargs)

        self._header = None

    @property
    def header(self):
        if "pixel_size" in self.meta:
            hdr = self.get_header(self.meta["pixel_size"])
        else:
            hdr = self._header

        return hdr

    def get_header(self, pixel_size):

        dx = self.x[-1] - self.x[0]     # [mm]
        dy = self.y[-1] - self.y[0]     # [mm]
        ds = self.s[-1] - self.s[0]     # [mm]
        len_mm = (dx**2 + dy**2)**0.5   # [mm]
        arcsec_per_mm = ds / len_mm     # [arcsec / mm]

        x_sign = 1 if dx * ds >= 0 else -1
        y_sign = 1 if dy * ds >= 0 else -1

        hdr_dict = {"NAXIS": 2,
                    "NAXIS1": int(round(len_mm / pixel_size)),
                    "NAXIS2": 0,
                    "CTYPE1D": "LINEAR",
                    "CTYPE2D": "LINEAR",
                    "CUNIT1D": "mm",
                    "CUNIT2D": "mm",
                    "CDELT1D": pixel_size * x_sign,    # [mm / pix]
                    "CDELT2D": pixel_size * y_sign,    # [mm / pix]
                    "CRVAL1D": self.x[0],
                    "CRVAL2D": self.y[0],
                    "CRPIX1D": 0,
                    "CRPIX2D": 0,
                    }

        rad2deg = np.pi / 180.
        c = np.cos(self.meta["rotation"] * rad2deg)
        s = np.sin(self.meta["rotation"] * rad2deg)
        t = np.tan(self.meta["shear"] * rad2deg)
        pc_dict = {"PC1_1D": c, "PC1_2D": s + c*t,
                   "PC2_1D": -s, "PC2_2D": c + s*t}
        hdr_dict.update(pc_dict)

        # ..todo: deal with these aspects
        # curvature

        # self._header = fits.Header(hdr_dict)
        self._header = PoorMansHeader(hdr_dict)
        self._header["WAVE_MIN"] = self.wave_min
        self._header["WAVE_MID"] = 0.5 * (self.wave_min + self.wave_max)
        self._header["WAVE_MAX"] = self.wave_max
        self._header["PLATESCL"] = (arcsec_per_mm,
                                    "[arcsec mm-1] Along slit")
        self._header["ROTANGD"] = (self.meta["rotation"] * rad2deg,
                                   "[deg] w.r.t. focal plane")
        self._header["SKEWANGD"] = (self.meta["shear"] * rad2deg,
                                    "[deg] w.r.t. trace trajectory")

        return self._header



import numpy as np

from astropy import units as u
from astropy.io import fits
from astropy.table import Table, Column

from . import fov_utils
from . import image_plane_utils as imp_utils

from ..base_classes import SourceBase, FieldOfViewBase, PoorMansHeader
from .. import utils


class FieldOfView(FieldOfViewBase):
    """
    A FOV is a monochromatic image. Flux units after extracting the fields from
    the Source are in ph/s/pixel

    The initial header should contain an on-sky WCS description:
    - CDELT-, CUNIT-, NAXIS- : for pixel scale and size (assumed CUNIT in deg)
    - CRVAL-, CRPIX- : for positioning the final image
    - CTYPE- : is assumed to be "RA---TAN", "DEC---TAN"

    and an image-plane WCS description
    - CDELT-D, CUNIT-D, NAXISn : for pixel scale and size (assumed CUNIT in mm)
    - CRVAL-D, CRPIX-D : for positioning the final image
    - CTYPE-D : is assumed to be "LINEAR", "LINEAR"

    The wavelength range is given by waverange

    """

    def __init__(self, header, waverange, **kwargs):
        self.meta = {"id": None,
                     "wave_min": utils.quantify(waverange[0], u.um),
                     "wave_max": utils.quantify(waverange[1], u.um),
                     "area": 0 * u.m**2,
                     "sub_pixel": "!SIM.sub_pixel.flag",
                     "distortion": {"scale": [1, 1],
                                    "offset": [0, 0],
                                    "shear": [1, 1],
                                    "rotation": 0,
                                    "radius_of_curvature": None},
                     "conserve_image": True,
                     "trace_id": None,
                     "aperture_id": None,
                     }
        self.meta.update(kwargs)

        if not any([utils.has_needed_keywords(header, s) for s in ["", "S"]]):
            raise ValueError("header must contain a valid sky-plane WCS: {}"
                             "".format(dict(header)))
        if not utils.has_needed_keywords(header, "D"):
            raise ValueError("header must contain a valid image-plane WCS: {}"
                             "".format(dict(header)))

        if isinstance(header, PoorMansHeader):
            self.hdu = fits.ImageHDU()
            self.hdu.header.update(header)
        else:
            self.hdu = fits.ImageHDU(header=header)
        self.hdu.header["NAXIS"] = 2
        self.hdu.header["NAXIS1"] = header["NAXIS1"]
        self.hdu.header["NAXIS2"] = header["NAXIS2"]

        self.image_plane_id = 0
        self.fields = []
        self.spectra = []

        self.cube = None        # IFU, long-lit, Slicer-MOS
        self.image = None       # Imagers
        self.spectrum = None    # Fibre-fed MOS

        self._waverange = None
        self._wavelength = None
        self._volume = None

    def extract_from(self, src):
        """ ..assumption: Bandpass has been applied"""

        if not isinstance(src, SourceBase):
            raise ValueError("source must be a Source object: {}"
                             "".format(type(src)))

        area = self.meta["area"]

        fields = [fov_utils.extract_common_field(field, self.volume())
                  for field in src.fields
                  if fov_utils.is_field_in_fov(self.hdu.header, field)]

        spec_refs = []
        for field in fields:
            if isinstance(field, Table):
                spec_refs += list(field["ref"])
            elif isinstance(field, fits.ImageHDU):
                ref = field.header.get("SPEC_REF")
                if ref is not None:
                    spec_refs += [ref]
        spectra = {ref: src.spectra[ref] for ref in spec_refs}

        self.fields = fields
        self.spectra = spectra


        ################### OLD ################################################
        #
        # # determine which fields are inside the field of view
        # fields_mask = [fov_utils.is_field_in_fov(self.hdu.header, field)
        #                for field in src.fields]
        # fields_indexes = np.where(fields_mask)[0]
        # tbl_fields_mask = np.array([isinstance(field, Table)
        #                             for field in src.fields])
        # img_fields_mask = np.array([isinstance(field, fits.ImageHDU)
        #                             for field in src.fields])
        #
        # # combine all Table fields
        # if sum(tbl_fields_mask * fields_mask) > 0:
        #     combined_table = fov_utils.combine_table_fields(self.hdu.header,
        #                                                     src, fields_indexes)
        #     tbl = fov_utils.make_flux_table(combined_table, src,
        #                                     wave_min, wave_max, area)
        #     xd, yd = fov_utils.sky2fp(self.hdu.header, tbl["x"], tbl["y"])
        #     tbl.add_columns([Column(name="x_mm", data=xd, unit=u.mm),
        #                      Column(name="y_mm", data=yd, unit=u.mm)])
        #     self.fields += [tbl]
        #
        # # combine all ImageHDU fields
        # if sum(img_fields_mask * fields_mask) > 0:
        #     imagehdu = fov_utils.combine_imagehdu_fields(self.hdu.header, src,
        #                                                  fields_indexes,
        #                                                  wave_min, wave_max,
        #                                                  area)
        #     self.fields += [imagehdu]

        ################### OLD ################################################

    def view(self, sub_pixel=None):
        return None

        # sum the cube layers within the wavelength range

        # combine images
        # add point sources to a canvas


        ################### OLD ################################################
        # if sub_pixel is None:
        #     sub_pixel = self.meta["sub_pixel"]
        #
        # self.hdu.data = np.zeros((  ))
        # if len(self.fields) > 0:
        #     for field in self.fields:
        #         if isinstance(field, Table):
        #             self.hdu = imp_utils.add_table_to_imagehdu(field, self.hdu,
        #                                                        sub_pixel)
        #         elif isinstance(field, fits.ImageHDU):
        #             self.hdu.data += field.data
        #
        # if self.meta["conserve_image"] is False and self.mask is not None:
        #     flux = np.sum(self.hdu.data) / np.sum(self.mask)
        #     self.hdu.data = np.zeros(self.hdu.data.shape)
        #     self.hdu.data[self.mask] = flux
        #
        # return self.hdu.data
        #
        # ################### OLD ################################################

    def make_spectrum(self):
        # This is needed for when we do incoherent MOS instruments.
        # Each fibre doesn't care about the spatial information.
        return None

    def make_image(self):
        # Used for imaging
        return None

    def make_cube(self):
        """
        Create a single cube out of all Source objects

        THE CURRENT IMPLEMENTATION IS TEMPORARY AND WORKS FOR A SINGLE
        CUBE SOURCE ONLY.
        """
        # Used for IFUs, slit spectrographs, and coherent MOSs (e.g.KMOS)

        ## The following is temporary [OC, 2021-07-27]
        ## Straight copy of fields[0], so that a cube is available.
        ## Works only when a single cube source is specified
        return self.fields[0]

    def volume(self, wcs_prefix=""):
        xs, ys = imp_utils.calc_footprint(self.header, wcs_suffix=wcs_prefix)
        wave_corners = self.waverange
        self._volume = {"xs": [min(xs), max(xs)],
                        "ys": [min(ys), max(ys)],
                        "waves": self.waverange,
                        "xy_unit": "mm" if wcs_prefix == "D" else "deg",
                        "wave_unit": "um"}
        return self._volume

    @property
    def header(self):
        return self.hdu.header

    @property
    def data(self):
        if self.hdu.data is None:
            self.view(self.meta["sub_pixel"])
        return self.hdu.data

    @property
    def corners(self):
        sky_corners = imp_utils.calc_footprint(self.header)
        imp_corners = imp_utils.calc_footprint(self.header, "D")
        return sky_corners, imp_corners

    @property
    def waverange(self):
        """Returns wavelength range in um [wave_min, wave_max]"""
        if self._waverange is None:
            wave_min = utils.quantify(self.meta["wave_min"], u.um).value
            wave_max = utils.quantify(self.meta["wave_max"], u.um).value
            self._waverange = [wave_min, wave_max]
        return self._waverange

    @property
    def wavelength(self):
        """Returns central wavelength in um"""
        if self._wavelength is None:
            self._wavelength = np.average(self.waverange)
        return self._wavelength

    def __repr__(self):
        msg = "FOV id: {}, with dimensions ({}, {})\n" \
              "".format(self.meta["id"], self.header["NAXIS1"],
                        self.header["NAXIS2"])
        msg += "Sky centre: ({}, {})\n" \
               "".format(self.header["CRVAL1"], self.header["CRVAL2"])
        msg += "Image centre: ({}, {})\n" \
               "".format(self.header["CRVAL1D"], self.header["CRVAL2D"])
        msg += "Wavelength range: ({}, {})um\n" \
               "".format(self.meta["wave_min"], self.meta["wave_max"])

        return msg



"""
# old functionality to implement:
# - provide x, y, lam, spectra, weight, ref
# - overridden + : number, Source, SourceSpectrum
# - overridden * : number, SpectralElement
# - write to and read from file
# - shift all fields
# - rotate around the centre
# - photons_in_range returns the photons per spectrum in a wavelength range
# - image_in_range returns an image of the source for a wavelength range
#
# old functionality which will be removed:
# - project_onto_chip
# - apply_optical_train
#
# old structure --> new structure:
# - all data held in 6 arrays
# --> new dicts for fields, spectrum
#       field can be a Table or an ImageHDU
#       spectrum is a SourceSpectrum
#
# Use cases:
# image + spectrum
# images + spectra
# table + spectrum
# table + spectra
#
# table columns = x, y, spec_id, weight
# table meta keywords = x_unit, y_unit
#
# image header keywords = WCS, SPEC_ID, WEIGHT
# [WCS = CRPIXn, CRVALn = (0,0), CTYPEn, CDn_m, NAXISn, CUNITn
"""

import os
import pickle
import logging
from copy import deepcopy
import numpy as np

from astropy.table import Table, Column
from astropy.io import ascii as ioascii
from astropy.io import fits
from astropy import units as u
from astropy.wcs import WCS

from synphot import SpectralElement, SourceSpectrum, Empirical1D
from synphot.units import PHOTLAM

from ..optics.image_plane import ImagePlane
from ..optics import image_plane_utils as imp_utils
from .source_utils import validate_source_input, convert_to_list_of_spectra, \
    photons_in_range
from . import source_templates as src_tmp

from ..base_classes import SourceBase
from .. import utils


class Source(SourceBase):
    """
    Create a source object from a file or from arrays

    A Source object must consist of a spatial and a spectral description
    of the on-sky source. Many sources can be added together and kept
    in memory as a single Source object.

    The spatial descriptions are kept in the ``<Source>.fields`` list,
    while the spectral descriptions are in the ``<Source>.spectra`` list.

    The spatial description can be built from any combination of:

    * a list of arrays (like in SimCADO >v0.5)
    * astropy Table objects
    * astropy ImageHDU objects
    * on disk FITS files
    * on disk ASCII tables

    The spectral descriptions can be passed as either ``synphot.SourceSpectrum``
    objects, or a set of two equal length arrays for wavelength and flux.

    .. hint:: Initialisation parameter combinations include:

        New ScopeSim-style input
        - ``table=<astropy.Table>, spectra=<list of synphot.SourceSpectrum>``
        - ``table=<astropy.Table>, lam=<array>, spectra=<list of array>``
        - ``image_hdu=<fits.ImageHDU>, spectra=<list of synphot.SourceSpectrum>``
        - ``image_hdu=<fits.ImageHDU>, lam=<array>, spectra=<list of array>``
        - ``image_hdu=<fits.ImageHDU>, flux=<astropy.Quantity>``

        Old SimCADO-style input
        - ``x=<array>, y=<array>, ref=<array>, spectra=<list of synphot.SourceSpectrum>``
        - ``x=<array>, y=<array>, ref=<array>, spectra=<list of array>, lam=<array>``
        - ``x=<array>, y=<array>, ref=<array>, weight=<array>, spectra=<list of array>, lam=<array>``

        More details on the content of these combinations can be found in the
        use-case documentation.

    Parameters
    ----------
    filename : str

    lam : np.array
        [um] Wavelength bins of length (m)
    spectra : list of synphot.SourceSpectra
        [ph/s/cm2/AA]
    x, y : np.array
        [arcsec] coordinates of where the emitting files are relative to the
        centre of the field of view
    ref : np.array
        the index for .spectra which connects a position (x, y) to a spectrum
        ``flux(x[i], y[i]) = spectra[ref[i]] * weight[i]``
    weight : np.array
        A weighting to scale the relevant spectrum for each position
    table : astropy.Table
    image_hdu : fits.ImageHDU
        [arcsec-2] The .data array is simply a map of weights for the assiciated
        spectrum referenced by .header["SPEC_REF].
        Surface brightness values are assumed to be per arcsec2
    flux : astropy.Quantity
        [u.mag, u.ABmag, u.Jy] Flux values are converted to a reference spectrum
        that is referenced by image_hdu.header["SPEC_REF"].
        flux can only be used in conjuction with image_hdu

    Attributes
    ----------
    fields : list
        The spatial distribution of the on-sky source, either as
        ``fits.ImageHDU`` or ``astropy.Table`` objects
    spectra : list of ``synphot.SourceSpectrum`` objects
        List of spectra associated with the fields
    meta : dict
        Dictionary of extra information about the source

    See Also
    --------
    synphot : ``https://synphot.readthedocs.io/en/latest/``

    """

    def __init__(self, filename=None, cube=None, ext=0,
                 lam=None, spectra=None, x=None, y=None, ref=None, weight=None,
                 table=None, image_hdu=None, flux=None, **kwargs):

        self.meta = {}
        self.meta.update(kwargs)
        self._meta_dicts = [self.meta]

        self.fields = []
        self.spectra = []

        self.bandpass = None

        validate_source_input(lam=lam, x=x, y=y, ref=ref, weight=weight,
                              spectra=spectra, table=table, cube=cube,
                              ext=ext, image_hdu=image_hdu, flux=flux,
                              filename=filename)

        if spectra is not None:
            spectra = convert_to_list_of_spectra(spectra, lam)

        if filename is not None and spectra is not None:
            self._from_file(filename, spectra, flux)

        elif cube is not None:
            self._from_cube(cube=cube, ext=ext)

        elif table is not None and spectra is not None:
            self._from_table(table, spectra)

        elif image_hdu is not None and spectra is not None:
            self._from_imagehdu_and_spectra(image_hdu, spectra)

        elif image_hdu is not None and flux is not None:
            self._from_imagehdu_and_flux(image_hdu, flux)

        elif image_hdu is not None and flux is None and spectra is None:
            if image_hdu.header.get("BUNIT") is not None:
                self._from_imagehdu_only(image_hdu)
            else:
                msg = f"image_hdu must be accompanied by either spectra or flux:\n" \
                      f"spectra: {spectra}, flux: {flux}"
                logging.exception(msg)
                raise ValueError(msg)

        elif x is not None and y is not None and \
                ref is not None and spectra is not None:
            self._from_arrays(x, y, ref, weight, spectra)

    def _from_file(self, filename, spectra, flux):
        filename = utils.find_file(filename)

        if utils.is_fits(filename):
            fits_type = utils.get_fits_type(filename)
            data = fits.getdata(filename)
            hdr = fits.getheader(filename)
            hdr['FILENAME'] = os.path.basename(filename)
            if fits_type == "image":
                image = fits.ImageHDU(data=data, header=hdr)
                if spectra is not None:
                    self._from_imagehdu_and_spectra(image, spectra)
                elif flux is not None:
                    self._from_imagehdu_and_flux(image, flux)
                else:
                    self._from_imagehdu_only(image)
            elif fits_type == "bintable":
                hdr1 = fits.getheader(filename, 1)
                hdr.update(hdr1)
                tbl = Table(data, meta=dict(hdr))
                tbl.meta.update(utils.convert_table_comments_to_dict(tbl))
                self._from_table(tbl, spectra)
        else:
            tbl = ioascii.read(filename)
            tbl.meta.update(utils.convert_table_comments_to_dict(tbl))
            self._from_table(tbl, spectra)

    def _from_table(self, tbl, spectra):
        if "weight" not in tbl.colnames:
            tbl.add_column(Column(name="weight", data=np.ones(len(tbl))))
        tbl["ref"] += len(self.spectra)
        self.fields += [tbl]
        self.spectra += spectra

    def _from_imagehdu_and_spectra(self, image_hdu, spectra):
        if not image_hdu.header.get("BG_SRC"):
            image_hdu.header["CRVAL1"] = 0
            image_hdu.header["CRVAL2"] = 0
            image_hdu.header["CRPIX1"] = image_hdu.header["NAXIS1"] / 2
            image_hdu.header["CRPIX2"] = image_hdu.header["NAXIS2"] / 2
            #image_hdu.header["CRPIX1"] = (image_hdu.header["NAXIS1"] + 1) / 2
            #image_hdu.header["CRPIX2"] = (image_hdu.header["NAXIS2"] + 1) / 2
            # .. todo:: find where the actual problem is with negative CDELTs
            # .. todo:: --> abs(pixel_scale) in header_from_list_of_xy
            if image_hdu.header["CDELT1"] < 0:
                image_hdu.header["CDELT1"] *= -1
                image_hdu.data = image_hdu.data[:, ::-1]
            if image_hdu.header["CDELT2"] < 0:
                image_hdu.header["CDELT2"] *= -1
                image_hdu.data = image_hdu.data[::-1, :]

        if isinstance(image_hdu, fits.PrimaryHDU):
            image_hdu = fits.ImageHDU(data=image_hdu.data,
                                      header=image_hdu.header)

        if spectra is not None and len(spectra) > 0:
            image_hdu.header["SPEC_REF"] = len(self.spectra)
            self.spectra += spectra
        else:
            image_hdu.header["SPEC_REF"] = ""
            logging.warning("No spectrum was provided. SPEC_REF set to ''. "
                            "This could cause problems later")
            raise NotImplementedError

        for i in [1, 2]:
            # Do not test for CUNIT or CDELT so that it throws an exception
            unit = u.Unit(image_hdu.header["CUNIT"+str(i)].lower())
            val = float(image_hdu.header["CDELT"+str(i)])
            image_hdu.header["CUNIT"+str(i)] = "DEG"
            image_hdu.header["CDELT"+str(i)] = val * unit.to(u.deg)

        self.fields += [image_hdu]

    def _from_imagehdu_and_flux(self, image_hdu, flux):
        if isinstance(flux, u.Unit):
            flux = 1 * flux

        spec_template = src_tmp.vega_spectrum
        if isinstance(flux, u.Quantity):
            if flux.unit.physical_type == "spectral flux density":  # ABmag and Jy
                spec_template = src_tmp.ab_spectrum
                flux = flux.to(u.ABmag)
            flux = flux.value
        spectra = [spec_template(flux)]
        self._from_imagehdu_and_spectra(image_hdu, spectra)

    def _from_imagehdu_only(self, image_hdu):
        bunit = image_hdu.header.get("BUNIT")
        try:
            bunit = u.Unit(bunit)
        except ValueError:
            f"Astropy cannot parse BUNIT [{bunit}].\n" \
            f"You can bypass this check by passing an astropy Unit to the flux parameter:\n" \
            f">>> Source(image_hdu=..., flux=u.Unit(bunit), ...)"

        value = 0 if bunit in [u.mag, u.ABmag] else 1
        self._from_imagehdu_and_flux(image_hdu, value * bunit)

    def _from_arrays(self, x, y, ref, weight, spectra):
        if weight is None:
            weight = np.ones(len(x))

        x = utils.quantify(x, u.arcsec)
        y = utils.quantify(y, u.arcsec)
        tbl = Table(names=["x", "y", "ref", "weight"],
                    data=[x, y, np.array(ref) + len(self.spectra), weight])
        tbl.meta["x_unit"] = "arcsec"
        tbl.meta["y_unit"] = "arcsec"

        self.fields += [tbl]
        self.spectra += spectra

    def _from_cube(self, cube, ext=0):
        """

        Parameters
        ----------
        cube: a file, HDUList or a PrimaryHDU object containing the cube
        ext: int
            the extension where the cube is located if applicable.

        """
        if isinstance(cube, fits.HDUList):
            data = cube[ext].data
            header = cube[ext].header
            wcs = WCS(cube[ext], fobj=cube)
        elif isinstance(cube, (fits.PrimaryHDU, fits.ImageHDU)):
            data = cube.data
            header = cube.header
            wcs = WCS(cube)
        else:
            with fits.open(cube) as hdul:
                data = hdul[ext].data
                header = hdul[ext].header
                header['FILENAME'] = os.path.basename(cube)
                wcs = WCS(cube)

        try:
            bunit = header['BUNIT']
            u.Unit(bunit)
        except KeyError:
            bunit = "erg / (s cm2 arcsec2)"
            logging.warning("Keyword 'BUNIT' not found, setting to %s by default", bunit)
        except ValueError as errcode:
            print("'BUNIT' keyword is malformed:", errcode)
            raise

        # Compute the wavelength vector. This will be attached to the cube_hdu
        # as a new `wave` attribute.  This is not optimal coding practice.
        wave = wcs.all_pix2world(header['CRPIX1'], header['CRPIX2'],
                                 np.arange(data.shape[0]), 0)[-1]

        wave = (wave * u.Unit(wcs.wcs.cunit[-1])).to(u.um,
                                                     equivalencies=u.spectral())

        # WCS keywords must be updated because astropy.wcs converts wavelengths to 'm'
        header.update(wcs.to_header())

        target_cube = data
        target_hdr = header.copy()
        target_hdr["BUNIT"] = bunit

        cube_hdu = fits.ImageHDU(data=target_cube, header=target_hdr)
        cube_hdu.wave = wave          # ..todo: review wave attribute, bad practice

        self.fields += [cube_hdu]

    @property
    def table_fields(self):
        """List of fields that are defined through tables"""
        fields = [field for field in self.fields if isinstance(field, Table)]
        return fields

    @property
    def image_fields(self):
        """List of fields that are defined through two-dimensional images"""
        fields = [field for field in self.fields if
                  isinstance(field, fits.ImageHDU) and field.header["NAXIS"] == 2]
        return fields

    @property
    def cube_fields(self):
        """List of fields that are defined through three-dimensional cubes"""
        fields = [field for field in self.fields if
                  isinstance(field, fits.ImageHDU) and field.header["NAXIS"] == 3]
        return fields

    # ..todo: rewrite this method
    def image_in_range(self, wave_min, wave_max, pixel_scale=1*u.arcsec,
                       layers=None, area=None, spline_order=1, sub_pixel=False):
        if layers is None:
            layers = range(len(self.fields))
        fields = [self.fields[ii] for ii in layers]

        hdr = imp_utils.get_canvas_header(fields, pixel_scale=pixel_scale)
        im_plane = ImagePlane(hdr)

        for field in fields:
            if isinstance(field, Table):
                fluxes = self.photons_in_range(wave_min, wave_max, area,
                                               field["ref"]) * field["weight"]
                x = utils.quantity_from_table("x", field, u.arcsec)
                y = utils.quantity_from_table("y", field, u.arcsec)
                tbl = Table(names=["x", "y", "flux"], data=[x, y, fluxes])
                tbl.meta.update(field.meta)
                hdu_or_table = tbl

            elif isinstance(field, fits.ImageHDU):
                if field.header["SPEC_REF"] != "":
                    ref = [field.header["SPEC_REF"]]
                    flux = self.photons_in_range(wave_min, wave_max, area, ref)
                    # [ph s-1] or [ph s-1 m-2] come out of photons_in_range

                # ## ..todo: CATCH UNITS HERE. DEAL WITH THEM PROPERLY
                # Currently assuming that all images are scaled appropriately
                # and that they have SPEC_REF

                # else:
                #     field = scale_imagehdu(field, area=area,
                #                            solid_angle=pixel_scale**2,
                #                            waverange=(wave_min, wave_max))
                #     # [ph s-1] or [ph s-1 m-2] come out of photons_in_range
                #     flux = 1

                image = field.data * flux
                hdu = fits.ImageHDU(header=field.header, data=image)
                hdu_or_table = hdu
            else:
                continue

            im_plane.add(hdu_or_table, sub_pixel=sub_pixel,
                         spline_order=spline_order)

        return im_plane

    def photons_in_range(self, wave_min, wave_max, area=None, indexes=None):
        """

        Parameters
        ----------
        wave_min : float, u.Quantity
            [um]
        wave_max : float, u.Quantity
            [um]
        area : float, u.Quantity, optional
            [m2]
        indexes : list of integers, optional

        Returns
        -------
        counts : u.Quantity list
            [ph / s / m2] if area is None
            [ph / s] if area is passed

        """
        if indexes is None:
            indexes = range(len(self.spectra))

        spectra = [self.spectra[ii] for ii in indexes]
        counts = photons_in_range(spectra, wave_min, wave_max, area=area,
                                  bandpass=self.bandpass)
        return counts

    def fluxes(self, wave_min, wave_max, **kwargs):
        return self.photons_in_range(wave_min, wave_max, **kwargs)

    def image(self, wave_min, wave_max, **kwargs):
        return self.image_in_range(wave_min, wave_max, **kwargs)

    @classmethod
    def load(cls, filename):
        """Load :class:'.Source' object from filename"""
        with open(filename, 'rb') as fp1:
            src = pickle.load(fp1)
        return src

    def dump(self, filename):
        """Save to filename as a pickle"""
        with open(filename, 'wb') as fp1:
            pickle.dump(self, fp1)

    # def collapse_spectra(self, wave_min=None, wave_max=None):
    #     for spec in self.spectra:
    #         waves = spec.waveset
    #         if wave_min is not None and wave_max is not None:
    #             mask = (waves >= wave_min) * (waves <= wave_max)
    #             waves = waves[mask]
    #         fluxes = spec(waves)
    #         spec = SourceSpectrum(Empirical1D, points=waves,
    #                               lookup_table=fluxes)

    def shift(self, dx=0, dy=0, layers=None):
        """
        Shifts the position of one or more fields w.r.t. the optical axis

        Parameters
        ----------
        dx, dy : float
            [arcsec]
        layers : list of ints
            which .fields entries to shift

        """

        if layers is None:
            layers = np.arange(len(self.fields))

        for ii in layers:
            if isinstance(self.fields[ii], Table):
                x = utils.quantity_from_table("x", self.fields[ii], u.arcsec)
                x += utils.quantify(dx, u.arcsec)
                self.fields[ii]["x"] = x

                y = utils.quantity_from_table("y", self.fields[ii], u.arcsec)
                y += utils.quantify(dy, u.arcsec)
                self.fields[ii]["y"] = y
            elif isinstance(self.fields[ii], (fits.ImageHDU, fits.PrimaryHDU)):
                dx = utils.quantify(dx, u.arcsec).to(u.deg)
                dy = utils.quantify(dy, u.arcsec).to(u.deg)
                self.fields[ii].header["CRVAL1"] += dx.value
                self.fields[ii].header["CRVAL2"] += dy.value

    def rotate(self, angle, offset=None, layers=None):
        pass

    def add_bandpass(self, bandpass):
        if not isinstance(bandpass, SpectralElement):
            raise ValueError("type(bandpass) must be synphot.SpectralElement")

        self.bandpass = bandpass

    def plot(self):
        """
        Plot the location of source components

        Source components instantiated from 2d or 3d ImageHDUs are represented by their
        spatial footprint. Source components instantiated from tables are shown as points.
        """
        # pylint: disable=import-outside-toplevel
        import matplotlib.pyplot as plt

        colours = "rgbcymk" * (len(self.fields) // 7 + 1)
        for col, field in zip(colours, self.fields):
            if isinstance(field, Table):
                plt.plot(field["x"], field["y"], col+".")
            elif isinstance(field, (fits.ImageHDU, fits.PrimaryHDU)):
                xpts, ypts = imp_utils.calc_footprint(field.header)
                xpts *= 3600   # Because ImageHDUs are always in CUNIT=DEG
                ypts *= 3600
                xpts = list(xpts) + [xpts[0]]
                ypts = list(ypts) + [ypts[0]]
                plt.plot(xpts, ypts, col)
                plt.xlabel("x [arcsec]")
                plt.ylabel("y [arcsec]")
        plt.gca().set_aspect("equal")

    def make_copy(self):
        new_source = Source()
        new_source.meta = deepcopy(self.meta)
        new_source._meta_dicts = deepcopy(self._meta_dicts)
        new_source.spectra = deepcopy(self.spectra)
        for field in self.fields:
            if isinstance(field, (fits.ImageHDU, fits.PrimaryHDU)) \
                    and field._file is not None:  # and field._data_loaded is False:
                new_source.fields += [field]
            else:
                new_source.fields += [deepcopy(field)]

        return new_source

    def append(self, source_to_add):
        new_source = source_to_add.make_copy()
        if isinstance(source_to_add, Source):
            for field in new_source.fields:
                if isinstance(field, Table):
                    field["ref"] += len(self.spectra)
                    self.fields += [field]

                elif isinstance(field, (fits.ImageHDU, fits.PrimaryHDU)):
                    if ("SPEC_REF" in field.header and
                        isinstance(field.header["SPEC_REF"], int)):
                        field.header["SPEC_REF"] += len(self.spectra)
                    self.fields += [field]
                self.spectra += new_source.spectra

                self._meta_dicts += source_to_add._meta_dicts
        else:
            raise ValueError("Cannot add {} object to Source object"
                             "".format(type(new_source)))

    def __add__(self, new_source):
        self_copy = self.make_copy()
        self_copy.append(new_source)
        return self_copy

    def __radd__(self, new_source):
        return self.__add__(new_source)

    def __repr__(self):
        msg = ""
        for ifld, fld in enumerate(self.fields):
            if isinstance(fld, Table):
                tbl_len = len(fld)
                num_spec = set(fld["ref"])
                msg += f"[{ifld}]: Table with {tbl_len} rows, referencing spectra {num_spec} \n"
            elif isinstance(fld, (fits.ImageHDU, fits.PrimaryHDU)):
                im_size = fld.data.shape if fld.data is not None else "<empty>"
                num_spec = "-"
                msg += f"[{ifld}]: ImageHDU with size {im_size}"
                if "SPEC_REF" in self.fields[ifld].header:
                    num_spec = self.fields[ifld].header["SPEC_REF"]
                    msg += f", referencing spectrum {num_spec}"
                msg += "\n"

        return msg



import os

from astropy.table import TableFormatter
from docutils.core import publish_doctree, publish_parts
from docutils.nodes import comment, literal_block
import yaml

from .. import rc
from ..utils import from_currsys


def walk(node, context_code=None):
    """
    Recursively walk through a docutils doctree and run/plot code blocks

    Parameters
    ----------
    node : docutils.node.Node
    context_code : str, optional
        A code string inherited from previous code/comment nodes

    Returns
    -------
    context_code : str
        Code to be inherited by subsequent code/comment nodes

    """

    if context_code is None:
        context_code = """
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.colors import LogNorm
"""

    if isinstance(node, (comment, literal_block)):
        if isinstance(node, comment):
            context_code = process_comment_code(node, context_code)
        elif isinstance(node, literal_block):
            context_code = process_literal_code(node, context_code)

        exec(context_code)

    elif hasattr(node, "children"):
        for child in node.children:
            context_code = walk(child, context_code)

    return context_code


def process_comment_code(node, context_code):
    """
    Add code from a ``comment`` node to the context_code string
    """
    result = node.rawsource.split("---")
    options = yaml.full_load(result[0]) if len(result) > 1 else {}
    new_code = result[-1]
    context_code = process_code(context_code, new_code, options)

    return context_code


def process_literal_code(node, context_code):
    """Add code from a ``literal_block`` node to the context_code string"""
    new_code = node.rawsource
    attribs = node.attributes["classes"]
    action = [att for att in attribs if "format-" not in att]
    format = [att.replace("format-", "") for att in attribs if "format-" in att]
    format = format if len(format) > 0 else ["png"]
    options = {"action": action,
               "format": format}
    if len(node.attributes["names"]) > 0:
        options["name"] = node.attributes["names"][0]
    context_code = process_code(context_code, new_code, options)

    return context_code


def process_code(context_code, code, options):
    """
    Extracts and adds code from the node text to the context_code string

    Code can be passed in either a ``literal_block`` or ``comment`` docutils
    node. The options regarding what to do with the code are included in either
    the :class: and :name: tag of a ``literal_block`` node, or in a yaml header
    section in a ``comment`` node.

    See below for examples of code blocks.

    Options for controlling what happens to the code block are as follows:

    - name: The filename for the plot

    - format: Any matplotlib-accepted file format, e.g: png, pdf, svg, etc.
        Multiple file formats can be

    - action: [reset, clear-figure, plot]
        - ``reset`` clears the context_code string. By default code is saved
          from previous code blocks.
        - ``clear-figure`` adds ``plt.clf()`` to the context string before the
          current code block is added
        - ``plot`` adds ``plt.savefig({name}.{format})`` to the context string.
          If multiple formats are given, these will be iterated over.

    For ``comment`` blocks, options should be given in a yaml style header,
    separated form the code block by exactly three (3) hyphens ('---')

    For ``literal_block`` blocks, we hijack the ``:class:`` and ``:name:``
    attributes. See the examples below. All action keywords are passed to
    ``:class:``. Format keys are passed as ``format-<key>``.

    Parameters
    ----------
    context_code : str
        code from previous Nodes
    code : str
        code from the current Node
    options : dict
        options dictionary derived from Node attributes or RST header blocks

    Returns
    -------
    context_code : str

    Examples
    --------

    Example of ``literal_block`` code block::

        .. code::
            :name: my_fug
            :class: reset, clear-figure, plot, format-png

            plt.plot([0,1], [1,1])

        .. figure:: my_fug.png
            :name: fig:my_fug

    Example of a ``comment`` code block::

        ..
            name: my_fug2
            format: [jpg, svg]
            action: [reset, clear-figure, plot]
            ---
            plt.plot([0,1], [1,0])

        .. figure:: my_fug2.jpg
            :name: fig:my_fug2

    """

    if "reset" in options.get("action", []):
        context_code = ""

    if "clear-figure" in options.get("action", []):
        context_code += "\nimport matplotlib.pyplot as plt\nplt.clf()\n"

    if "execute" in options.get("action", []):
        context_code += code

    if "plot" in options.get("action", []):
        context_code += code

        img_path = options.get("path", rc.__config__["!SIM.reports.image_path"])
        formats = options.get("format", ["png"])
        formats = [formats] if isinstance(formats, str) else formats
        for fmt in formats:

            fname = options.get("name", "untitled").split(".")[0]
            fname = ".".join([fname, fmt])
            fname = os.path.join(img_path, fname)
            context_code += '\nplt.savefig("{}")'.format(fname)

    return context_code


def plotify_rst_text(rst_text):
    """
    Generates and saves plots from code blocks in an RST string

    The save directory for the plots defaults to
    ``scopesim.rc.__config__["!SIM.reports.image_path"]``.
    This can be overridden ONLY inside a COMMENT block using the path keyword.

    Parameters
    ----------
    rst_text : str
        Any RST text string

    Examples
    --------
    The following rst text will generate a plot and save it in three formats::

        A basic plot
        ============
        Let's make a basic plot using the comment style

        ..
            action: plot
            format: [pdf, png]
            name: my_fug
            path: "./images/"
            ---
            plt.plot([0,1], [0,1])

        And now a plot using the code block style

        .. code::
            :name: my_fug3
            :class: reset, plot, format-jpg, format-svg

            import matplotlib.pyplot as plt
            plt.plot([0,1], [1,1])

    The plot are created be calling::

        plotify_rst_text(rst_text)

    where ``rst_text`` is a string holding the full RST text from above.

    Notes
    -----
    * Possible actions are: [reset, clear-figure, plot]
    * Code is retained between code blocks in the same string, so we do not need
      to re-write large sections of code
    * By default ``import numpy as np`` and ``import matplotlib.pyplot as plt``
      are loaded automatically, so these do not need to be explicitly specified
      in each code block.
    * THE EXCEPTION is when the action ``reset`` is specified. This clears the
      code_context variable.

    """

    walk(publish_doctree(rst_text))


def latexify_rst_text(rst_text, filename=None, path=None, title_char="=",
                      float_figures=True, use_code_box=True):
    """
    Converts an RST string (block of text) into a LaTeX string

    NOTE: plots will NOT be generated with this command. For that we must invoke
    the ``plotfiy`` command.

    Parameters
    ----------
    rst_text : str
    filename : str, optional
        What to name the latex file. If None, the filename is derived from the
        text title
    path : str, optional
        Where to save the latex file
    title_char : str, optional
        The character used to underline the rst text title. Usually "=".
    float_figures : bool, optional
        Set to False if figures should not be placed by LaTeX.
        Replaces all ``\begin{figure}`` with ``\begin{figure}[H]``
    use_code_box : bool, optional
        Adds a box around quote blocks

    Returns
    -------
    tex_str : str
        The same string or block of text in LaTeX format

    Examples
    --------
    ::
        rst_text = '''
        Meaning of life
        ===============

        Apparently it's 42. Lets plot

        ..
            action: plot
            name: my_plot
            path: ./images/
            ---
            plt.plot([0,1], [0,1])

        .. figure:: images/my_plot.png
            :name: label-my-plot

            This is an included figure caption

        '''

        plotify_rst_text(rst_text)
        latexify_rst_text(rst_text, filename="my_latex_file", path="./")

    """

    if path is None:
        path = from_currsys(rc.__config__["!SIM.reports.latex_path"])

    if filename is None:
        filename = rst_text.split(title_char)[0].strip().replace(" ", "_")

    text = "Title\n<<<<<\nSubtitle\n>>>>>>>>\n\n"
    parts = publish_parts(text + rst_text, writer_name="latex")

    if not float_figures:
        parts["body"] = parts["body"].replace('begin{figure}',
                                              'begin{figure}[H]')

    if use_code_box:
        parts["body"] = parts["body"].replace('begin{alltt}',
                                              'begin{alltt}\n\\begin{lstlisting}[frame=single]')
        parts["body"] = parts["body"].replace('end{alltt}',
                                              'end{lstlisting}\n\\end{alltt}')

    filename = filename.split(".")[0] + ".tex"
    file_path = os.path.join(path, filename)
    with open(file_path, "w") as f:
        f.write(parts["body"])

    tex_str = parts["body"]

    return tex_str


def rstify_rst_text(rst_text, filename=None, path=None, title_char="="):
    """ The same as ``latexify_rst_text```, but the output is in RST format """
    if path is None:
        path = from_currsys(rc.__config__["!SIM.reports.rst_path"])

    if filename is None:
        filename = rst_text.split(title_char)[0].strip().replace(" ", "_")

    filename = filename.split(".")[0] + ".rst"
    file_path = os.path.join(path, filename)
    with open(file_path, "w") as f:
        f.write(rst_text)

    return rst_text


def table_to_rst(tbl, indent=0, rounding=None):
    if isinstance(rounding, int):
        for col in tbl.itercols():
            if col.info.dtype.kind == 'f':
                col.info.format = '.{}f'.format(rounding)
    
    tbl_fmtr = TableFormatter()
    lines, outs = tbl_fmtr._pformat_table(tbl, max_width=-1, max_lines=-1,
                                          show_unit=False)
    i = outs["n_header"] - 1
    lines[i] = lines[i].replace("-", "=")
    lines = [lines[i]] + lines + [lines[i]]

    indent = " " * indent
    rst_str = indent + ("\n" + indent).join(lines)

    return rst_str



import os
import numpy as np
from scopesim.effects import effects_utils as eu


def setup(app):
    output_dir = os.path.abspath(os.path.join(os.path.dirname(__file__),
                                 "../effects_docstrings"))
    os.makedirs(output_dir, exist_ok=True)

    efs_dict = eu.scopesim_effect_classes()
    eff_types = np.unique([eff.split(".")[0] for eff in efs_dict])
    eff_type_strs = {eff_type: f"{eff_type}\n{'='*len(eff_type)}\n\n"
                     for eff_type in eff_types}

    for eff, cls in efs_dict.items():
        new_str = f"{cls.__name__}\n{'+'*len(cls.__name__)}\n\n"
        if cls.__doc__ is not None:
            new_str += cls.__doc__.replace("\n    ", "\n") + "\n\n"
        else:
            new_str += ".. warning: Empty docstring\n\n"

        base_name = eff.split(".")[0]
        eff_type_strs[base_name] += new_str

    for eff_type, strs in eff_type_strs.items():
        with open(os.path.join(output_dir, f"{eff_type}.rst"), "w") as f:
            f.write(strs)


    toc_str = """Effect Descriptions
===================

.. toctree::
    :maxdepth: 2
    :caption: Contents:
    :glob:

    *
    
"""

    with open(os.path.join(output_dir, "index.rst"), "w") as f:
        f.write(toc_str)



import os
import yaml

from .system_dict import SystemDict

__pkg_dir__ = os.path.dirname(__file__)

with open(os.path.join(__pkg_dir__, "defaults.yaml")) as f:
    dicts = [dic for dic in yaml.full_load_all(f)]

user_rc_path = os.path.expanduser("~/.scopesim_rc.yaml")
if os.path.exists(user_rc_path):
    with open(user_rc_path) as f:
        dicts += [dic for dic in yaml.full_load_all(f)]

__config__ = SystemDict(dicts)
__currsys__ = __config__

__search_path__ = [__config__["!SIM.file.local_packages_path"],
                   __pkg_dir__] + __config__["!SIM.file.search_path"]

# if os.environ.get("READTHEDOCS") == "True" or "F:" in os.getcwd():
#     extra_paths = ["../", "../../", "../../../", "../../../../"]
#     __search_path__ = extra_paths + __search_path__


import os
import logging

import numpy as np

from astropy import units as u
from astropy.io import ascii as ioascii
from astropy.table import Table

from synphot import SpectralElement
from synphot.models import Empirical1D

from ..effects import ter_curves_utils as ter_utils
from ..utils import get_meta_quantity, quantify, extract_type_from_unit
from ..utils import from_currsys, convert_table_comments_to_dict, find_file
from .surface_utils import make_emission_from_emissivity,\
    make_emission_from_array


class PoorMansSurface:
    """ Solely used by SurfaceList """
    def __init__(self, emission, throughput, meta):
        self.emission = emission
        self.throughput = throughput
        self.meta = meta


class SpectralSurface:
    """
    Initialised by a file containing one or more of the following columns:
    transmission, emissivity, reflection. The column wavelength must be given.
    Alternatively kwargs for the above mentioned quantities can be passed as
    arrays. If they are not Quantities, then a unit should also be passed with
    the <array_name>_unit syntax (i.e. emission_unit or wavelength_unit)

    If temperature is not given as a Quantity, it defaults to degrees Celsius.

    """
    def __init__(self, filename=None, **kwargs):
        filename = find_file(filename)
        self.meta = {"filename"         : filename,
                     "temperature"      : -270*u.deg_C,  # deg C
                     "emission_unit"    : "",
                     "wavelength_unit"  : u.um}

        self.table = Table()
        if filename is not None and os.path.exists(filename):
            self.table = ioascii.read(filename)
            tbl_meta = convert_table_comments_to_dict(self.table)
            if isinstance(tbl_meta, dict):
                self.meta.update(tbl_meta)

        self.meta.update(kwargs)

    @property
    def area(self):
        if "area" in self.meta:
            the_area = self.from_meta("area", u.m**2)
        elif "outer" in self.meta:
            outer_diameter = self.from_meta("outer", u.m)
            the_area = np.pi * (0.5 * outer_diameter)**2
            if "inner" in self.meta:
                inner_diameter = self.from_meta("inner", u.m)
                the_area -= np.pi * (0.5 * inner_diameter) ** 2
        else:
            the_area = None

        return the_area

    @property
    def mirror_angle(self):
        if "angle" in self.meta:
            mirr_angle = self.from_meta("angle", u.deg)
        else:
            mirr_angle = 0 * u.deg
        return mirr_angle

    @property
    def wavelength(self):
        return self._get_array("wavelength")

    @property
    def throughput(self):
        return self._get_ter_property(self.meta.get("action", "transmission"))

    @property
    def transmission(self):
        return self._get_ter_property("transmission")

    @property
    def emissivity(self):
        return self._get_ter_property("emissivity")

    @property
    def reflection(self):
        return self._get_ter_property("reflection")

    @property
    def emission(self):
        """
        Looks for an emission array in self.meta. If it doesn't find this, it
        defaults to creating a blackbody and multiplies this by the emissivity.
        Assumption is that self.meta["temperature"] is in deg_C, unless it is
        a u.Quantity with attached temperature unit.
        Return units are in PHOTLAM arcsec^-2, even though arcsec^-2 is not
        given
        """

        flux = self._get_array("emission")
        if flux is not None:
            wave = self._get_array("wavelength")
            flux = make_emission_from_array(flux, wave, meta=self.meta)
        elif "temperature" in self.meta:
            emiss = self.emissivity                     # SpectralElement [0..1]
            temp = from_currsys(self.meta["temperature"])
            if not isinstance(temp, u.Quantity):
                temp = quantify(temp, u.deg_C)
            temp = temp.to(u.Kelvin, equivalencies=u.temperature())
            flux = make_emission_from_emissivity(temp, emiss)
            flux.meta["temperature"] = temp
        else:
            flux = None

        has_solid_angle = extract_type_from_unit(flux.meta["solid_angle"],
                                                 "solid angle")[1] != u.Unit("")
        if flux is not None and has_solid_angle:
            conversion_factor = flux.meta["solid_angle"].to(u.arcsec ** -2)
            flux = flux * conversion_factor
            flux.meta["solid_angle"] = u.arcsec**-2
            flux.meta["history"] += ["Converted to arcsec-2: {}"
                                     "".format(conversion_factor)]

        if flux is not None and "rescale_emission" in self.meta:
            dic = from_currsys(self.meta["rescale_emission"])
            amplitude = dic["value"] * u.Unit(dic["unit"])
            filter_name = dic["filter_name"]
            if "filename_format" in dic:
                filter_name = dic["filename_format"].format(filter_name)
            flux = ter_utils.scale_spectrum(flux, filter_name, amplitude)

        return flux

    def from_meta(self, key, default_unit=None):
        """
        Converts a specific value in the meta dict to a ``Quantity``

        Parameters
        ----------
        key : str
            Which key to pull from self.meta
        default_unit : str, Unit
            In case self.meta doesn't contain a unit for the desired key

        Returns
        -------
        meta_quantity : Quantity

        """

        if default_unit is None:
            default_unit = ""
        meta_quantity = get_meta_quantity(self.meta, key, u.Unit(default_unit))

        return meta_quantity

    def _get_ter_property(self, ter_property, fmt="synphot"):
        """
        Looks for arrays for transmission, emissivity, or reflection

        Parameters
        ----------
        ter_property : str
            ``transmission``, ``emissivity``, ``reflection``
        fmt : str

        Returns
        -------
            response_curve : ``synphot.SpectralElement``

        """

        compliment_names = ["transmission", "emissivity", "reflection"]
        ii = np.where([ter_property == name for name in compliment_names])[0][0]
        compliment_names.pop(ii)

        wave = self._get_array("wavelength")
        value_arr = self._get_array(ter_property)
        if value_arr is None:
            value_arr = self._compliment_array(*compliment_names)

        if value_arr is not None and wave is not None and fmt == "synphot":
            response_curve = SpectralElement(Empirical1D, points=wave,
                                             lookup_table=value_arr)
        elif fmt == "array":
            response_curve = value_arr
        else:
            response_curve = None
            logging.warning("Both wavelength and {} must be set"
                          "".format(ter_property))

        return response_curve

    def _compliment_array(self, colname_a, colname_b):
        """
        Returns an complimentary array using: ``a + b + c = 1``

        E.g. ``Emissivity = 1 - (Transmission + Reflection)``

        Parameters
        ----------
        colname_a : str
            Name of the first TER property to look for
        colname_b
            Name of the second TER property to look for

        Returns
        -------
        actual : ``synphot.SpectralElement``
            Complimentary spectrum to those given

        """

        compliment_a = self._get_array(colname_a)
        compliment_b = self._get_array(colname_b)

        if compliment_a is not None and compliment_b is not None:
            actual = 1*compliment_a.unit - (compliment_a + compliment_b)
        elif compliment_a is not None and compliment_b is None:
            actual = 1*compliment_a.unit - compliment_a
        elif compliment_b is not None and compliment_a is None:
            actual = 1*compliment_b.unit - compliment_b
        elif compliment_b is None and compliment_a is None:
            actual = None

        return actual

    def _get_array(self, colname):
        """
        Looks for an array in either the self.meta or self.table attributes

        Order of search goes: 1. self.meta, 2. self.table

        Parameters
        ----------
        colname : str
            Array column (or key) name

        Returns
        -------
            val_out : array-like Quantity

        """

        if colname in self.meta:
            val = self.meta[colname]
        elif colname in self.table.colnames:
            val = self.table[colname].data
        else:
            logging.debug(f"{colname} not found in either '.meta' or '.table': "
                          f"[{self.meta.get('name', self.meta['filename'])}]")
            return None

        col_units = colname+"_unit"
        if isinstance(val, u.Quantity):
            units = val.unit
        elif col_units in self.meta:
            units = u.Unit(self.meta[col_units])
        else:
            units = u.Unit("")

        if isinstance(val, u.Quantity):
            val_out = val.to(units)
        elif isinstance(val, (list, tuple, np.ndarray)):
            val_out = val * units
        elif val is None:
            val_out = None
        else:
            raise ValueError("{} must be of type: Quantity, array, list, tuple"
                             "".format(colname))

        return val_out

    def __repr__(self):
        meta = self.meta
        name = meta["name"] if "name" in meta else meta["filename"]
        cols = "".join([col[0].upper() for col in self.table.colnames])
        msg = '<SpectralSurface> [{}] "{}"'.format(cols, name)

        return msg



import numpy as np
from astropy import units as u
from astropy.table import Table

from .effects import Effect
from ..utils import airmass2zendist, from_currsys, check_keys, quantify
from ..base_classes import FieldOfViewBase


class Shift3D(Effect):
    def __init__(self, **kwargs):
        super(Shift3D, self).__init__(**kwargs)
        params = {"z_order": [30, 230],
                  "report_plot_include": True,
                  "report_table_include": False,}
        self.meta.update(params)
        self.meta.update(kwargs)

    def apply_to(self, obj, **kwargs):
        return obj

    def fov_grid(self, which="shifts", **kwargs):
        if which == "shifts":
            col_names = ["wavelength", "dx", "dy"]
            waves, dx, dy = [self.get_table(**kwargs)[col] for col in col_names]
            return waves, dx, dy
        else:
            return None

    def get_table(self, **kwargs):
        if self.table is None:
            names = ["wavelength", "dx", "dy"]
            waves = from_currsys(["!SIM.spectral.wave_" + key
                                  for key in ("min", "mid", "max")])
            tbl = Table(names=names, data=[waves, [0] * 3, [0] * 3])
        else:
            tbl = self.table

        return tbl

    def plot(self):
        import matplotlib.pyplot as plt
        plt.gcf().clf()

        tbl = self.get_table()
        plt.scatter(x=tbl["dx"], y=tbl["dy"], c=tbl["wavelength"])
        plt.colorbar()
        plt.xlabel("dx [{}]".format(quantify(tbl["dx"], u.arcsec).unit))
        plt.ylabel("dy [{}]".format(quantify(tbl["dy"], u.arcsec).unit))
        plt.axvline(0, ls=":")
        plt.axhline(0, ls=":")
        # plt.gca().set_aspect("equal")

        return plt.gcf()


class AtmosphericDispersion(Shift3D):
    """
    Used to generate the wavelength bins based on shifts due to the atmosphere

    Doesn't contain an ``apply_to`` function, but provides information through
    the ``fov_grid`` function.

    Required Parameters
    -------------------
    airmass : float
        Recommended to use "!OBS.airmass" in the OBS properties
    temperature : float
        [degC] Recommended to use "!ATMO.temperature" in the ATMO properties
    humidity : float
        [0..1] Recommended to use "!ATMO.humidity" in the ATMO properties
    pressure : float
        [bar] Recommended to use "!ATMO.pressure" in the ATMO properties
    latitude : float
        [deg] Recommended to use "!ATMO.latitude" in the ATMO properties
    altitude
        [m] Recommended to use "!ATMO.altitude" in the ATMO properties
    pixel_scale
        [arcsec] Recommended to use "!INST.pixel_scale" in the INST properties

    Optional Parameters
    -------------------
    wave_min : float
        [um] Defaults to "!SIM.spectral.wave_min"
    wave_mid : float
        [um] Defaults to "!SIM.spectral.wave_mid"
    wave_max : float
        [um] Defaults to "!SIM.spectral.wave_max"
    sub_pixel_fraction : float
        [0..1] Defaults to "!SIM.sub_pixel.fraction"
    num_steps : int
        Default: 1000. Number of wavelength steps to use when interpolating the
        atmospheric dispersion curve

    """
    def __init__(self, **kwargs):
        super(AtmosphericDispersion, self).__init__(**kwargs)
        params = {"z_order": [231],
                  "wave_min": "!SIM.spectral.wave_min",
                  "wave_mid": "!SIM.spectral.wave_mid",
                  "wave_max": "!SIM.spectral.wave_max",
                  "sub_pixel_fraction": "!SIM.sub_pixel.fraction",
                  "num_steps": 1000,}
        self.meta.update(params)
        self.meta.update(kwargs)

        required_keys = ["airmass", "temperature", "humidity", "pressure",
                         "latitude", "altitude", "pupil_angle", "pixel_scale"]
        check_keys(self.meta, required_keys, action="error")

    def get_table(self, **kwargs):
        """
        Called by the fov_grid method of Shift3D

        Returns
        -------
        tbl : astropy.Table
            A table with the columns
            - waves : [um]
            - dx, dy : [arcsec]

        Notes
        -----
        Success! Returns the same values as:
        http://gtc-phase2.gtc.iac.es/science/astroweb/atmosRefraction.php
        """

        if len(kwargs) > 0:
            self.meta.update(kwargs)

        airmass = from_currsys(self.meta["airmass"])
        atmo_params = {"z0": airmass2zendist(airmass),
                       "temp": self.meta["temperature"],  # in degC
                       "rel_hum": self.meta["humidity"] * 100,  # in %
                       "pres": self.meta["pressure"] * 1000,  # in mbar
                       "lat": self.meta["latitude"],  # in deg
                       "h": self.meta["altitude"]}  # in m
        self.meta.update(atmo_params)
        params = from_currsys(self.meta)

        waves, shifts = get_pixel_border_waves_from_atmo_disp(**params)
        dx = shifts * np.sin(np.deg2rad(params["pupil_angle"]))
        dy = shifts * np.cos(np.deg2rad(params["pupil_angle"]))

        names = ["wavelength", "dx", "dy"]
        tbl = Table(names=names, data=[waves, dx, dy])

        return tbl


class AtmosphericDispersionCorrection(Shift3D):
    def __init__(self, **kwargs):
        """
        Alters the position on the detector for a FOV object (WCS_prefix="D")

        Only acts on FOVs during the main effects loop in OpticalTrain.
        For the sake of computational efficiency, the ADC can be instructed to
        counteract the atmospheric diffraction during the OpticalTrain setup
        phase, by passing the kwarg: ``quick_adc=True``

        Parameters
        ----------
        kwargs
        """
        super(AtmosphericDispersionCorrection, self).__init__(**kwargs)
        self.meta["z_order"] = [632]
        if "quick_adc" in self.meta and self.meta["quick_adc"] is True:
            self.meta["z_order"] += [232]
        if "efficiency" not in self.meta:
            self.meta["efficiency"] = 1
        self.apply_to_classes = FieldOfViewBase

        required_keys = ["airmass", "temperature", "humidity", "pressure",
                         "latitude", "altitude", "pupil_angle", "pixel_scale",
                         "wave_mid"]
        check_keys(self.meta, required_keys, action="error")

        if self.table is None:
            self.table = self.get_table()

    def apply_to(self, fov, **kwargs):
        # .. todo:: Currently applying shift with pixel_scale to CRPIX-D
        # .. todo:: Change this to be applying to CRVAL-D using plate_scale
        # get mid wavelength of fov
        # work out on-sky shift
        # work out pupil-plane shift
        # convert to pixel shifts
        # correct fovs CRPIXnD keys

        if isinstance(fov, self.apply_to_classes):
            self.meta = from_currsys(self.meta)
            atmo_params = {"z0":        airmass2zendist(self.meta["airmass"]),
                           "temp":      self.meta["temperature"],  # in degC
                           "rel_hum":   self.meta["humidity"] * 100,  # in %
                           "pres":      self.meta["pressure"] * 1000, # in mbar
                           "lat":       self.meta["latitude"],  # in deg
                           "h":         self.meta["altitude"]}  # in m
            fov_wave_mid = quantify(fov.wavelength, u.um).value
            # .. todo:: why are we comparing to shift_obs. Is this static?
            obs_wave_mid = self.meta["wave_mid"]
            shift_obs = atmospheric_refraction(lam=obs_wave_mid, **atmo_params)
            shift_fov = atmospheric_refraction(lam=fov_wave_mid, **atmo_params)
            shift_rel_arcsec = shift_fov - shift_obs
            shift_rel_pixel = shift_rel_arcsec / self.meta["pixel_scale"]

            # this assumes a one-to-one mapping of sky coords to detector coords
            # this won't work if there is appreciable distortion in the optics
            # .. todo:: think about this
            pup_ang = self.meta["pupil_angle"]
            dy_pix = shift_rel_pixel * np.cos(np.deg2rad(pup_ang))
            dx_pix = shift_rel_pixel * np.sin(np.deg2rad(pup_ang))

            fov.header["CRPIX1D"] += dx_pix
            fov.header["CRPIX2D"] += dy_pix

        return fov

    def fov_grid(self, which="shifts", **kwargs):
        kwargs.update(self.meta)
        if "quick_adc" in self.meta:
            ad = AtmosphericDispersion(**self.meta)
            waves, dx, dy = ad.fov_grid()
            dx *= -(1 - self.meta["efficiency"])
            dy *= -(1 - self.meta["efficiency"])
            return waves, dx, dy
        else:
            return None

    def plot(self):
        return None


def atmospheric_refraction(lam, z0=60, temp=0, rel_hum=60, pres=750,
                           lat=-24.5, h=3064):
    """Compute atmospheric refraction

    The function computes the angular difference between the apparent position
    of a star seen from the ground and its true position.

    Parameters
    ----------
    lam : float, np.ndarray
        [um] wavelength bin centres
    z0 : float, optional
        [deg] zenith distance. Default is 60 deg from zenith
    temp : float, optional
        [deg C] ground temperature. Default is 0 deg C
    rel_hum : float, optional
        [%] relative humidity. Default is 60%
    pres : float, optional
        [millibar] air pressure. Default is 750 mbar
    lat : float, optional
        [deg] latitude. Default set for Cerro Armazones: 24.5 deg South
    h : float, optional
        [m] height above sea level. Default is 3064 m

    Returns
    -------
    ang : float, np.ndarray
        [arcsec] angle between real position and refracted position

    References
    ----------
    See Stone 1996 and the review by S. Pedraz -
    http://www.caha.es/newsletter/news03b/pedraz/newslet.html
    """

    # need T, P, RH for Ps, Pw Pa
    T = 273.15 + temp

    Ps = -10474. + 116.43 * T - 0.43284 * T**2 + 0.0005384 * T**3
    Pw = Ps * rel_hum / 100.
    Pa = pres - Pw

    # need n0 for gamma
    sig = 1. / lam
    Da = (Pa / T) * (1. + Pa * (57.9E-8 - 0.0009325 / T + 0.25844 / T**2))
    Dw = (Pw / T) * (1. + Pw * (1. + 3.7E-4 * Pw) *
                     (-2.37321E-3 + 2.23366 / T - 710.792 / T**2 +
                      77514.1 / T**3))

    na = Da * (2371.34 + 683939.7 / (130. - sig**2) + 4547.3 / (38.9 - sig**2))
    nw = Dw * (6487.31 + 58.058 * sig**2 - 0.7115 * sig**4 + 0.08851 * sig**6)
    n0 = 1E-8 * (na + nw) + 1.

    # need gamma, kappa and beta for R
    g = n0 - 1.
    b = 0.001254 * (273.15 + temp) / 273.15
    k = 1. + 0.005302 * np.sin(np.deg2rad(lat))**2 \
        - 0.00000583 * np.sin(2. * np.deg2rad(lat))**2 - 0.000000315 * h

    R = k * g * (1 - b) * np.tan(np.deg2rad(z0)) \
        - k * g * (b - g/2.) * np.tan(np.deg2rad(z0))**3

    # the refraction is the side of a triangle, although the triangle
    # side makes an arc across the sky.
    # We want the angle that this triangle side is subtending
    # Using the small angle approximation (which is in radians),
    # we can get the angle of refraction

    ang = np.rad2deg(R) * 3600

    # return value is in arcsec
    return ang


def get_pixel_border_waves_from_atmo_disp(**kwargs):
    """

    Parameters
    ----------
    kwargs

    Returns
    -------

    """
    atmo_disp_dict = {key: kwargs[key] for key in ["z0", "temp", "rel_hum",
                                                   "pres", "lat", "h"]}
    wave_range = np.linspace(kwargs["wave_min"], kwargs["wave_max"],
                             kwargs["num_steps"])
    wave_mid = kwargs["wave_mid"]
    offset_mid = atmospheric_refraction(lam=wave_mid, **atmo_disp_dict)
    offset_ang = atmospheric_refraction(lam=wave_range, **atmo_disp_dict)
    offset_ang -= offset_mid

    # .. todo:: replace the 1e-7 with a variable in !SIM
    if np.any(np.abs(offset_ang) > 1e-7):
        offset_step = kwargs["pixel_scale"] * kwargs["sub_pixel_fraction"]
        offset_pix = offset_ang / offset_step

        # interpolate to get the edge wavelengths of the pixels
        # off_new is always increasing, thanks to np.unique
        off_new = np.unique(offset_pix.astype(int))
        # add 1 pixel to either end of the range covered by (wave_min, wave_max)
        off_new = np.array([off_new[0]-1] + list(off_new) + [off_new[-1]+1])

        if offset_pix[0] > offset_pix[-1]:
            wave_range = wave_range[::-1]
            offset_pix = offset_pix[::-1]

        wave_new = np.interp(off_new, offset_pix, wave_range)

        if wave_new[0] > wave_new[-1]:
            wave_new = wave_new[::-1]
            off_new = off_new[::-1]

        shifts_angle_edges = off_new * offset_step
        wave_pixel_edges = wave_new

    else:
        wave_pixel_edges = np.array([kwargs["wave_min"], kwargs["wave_max"]])
        shifts_angle_edges = np.zeros(2)

    return wave_pixel_edges, shifts_angle_edges



"""
Effect for mapping spectral cubes to the detector plane

The Effect is called SpectralTraceList, it applies a list of
optics.spectral_trace_SpectralTrace objects to a FieldOfView.
"""

from os import path as pth
import numpy as np

from astropy.io import fits
from astropy.table import Table

from .effects import Effect
from .spectral_trace_list_utils import SpectralTrace
from ..utils import from_currsys, check_keys, interp2
from ..optics.image_plane_utils import header_from_list_of_xy
from ..base_classes import FieldOfViewBase, FOVSetupBase


class SpectralTraceList(Effect):
    """
    List of spectral trace geometries for the detector plane

    Should work in concert with an ApertureList (or ApertureMask) object and a
    DetectorList object

    Spectral trace patterns are to be kept in a ``fits.HDUList`` with one or
    more ``fits.BinTableHDU`` extensions, each one describing the geometry of a
    single trace. The first extension should be a ``BinTableHDU`` connecting the
    traces to the correct ``Aperture`` and ``ImagePlane`` objects.

    The ``fits.HDUList`` objects can be loaded using one of these two keywords:

    - ``filename``: for on disk FITS files, or
    - ``hdulist``: for in-memory ``fits.HDUList`` objects

    The format and contents of the extensions in the HDUList (FITS file) object
    is listed below

    Input Data Format
    -----------------
    A trace list FITS file needs the following extensions:

    - 0 : PrimaryHDU [header]
    - 1 : BinTableHDU [header, data] : Overview table of all traces
    - 2..N : BinTableHDU [header, data] : Trace tables. One per spectral trace

    EXT 0 : PrimaryHDU
    ++++++++++++++++++
    Required Header Keywords:

    - ECAT : int : Extension number of overview table. Normally 1
    - EDATA : int : Extension number of first Trace table. Normally 2

    No data is required in this extension

    EXT 1 : BinTableHDU : Overview of traces
    ++++++++++++++++++++++++++++++++++++++++
    No special header keywords are required in this extension

    Required Table columns:

    - description : str : description of each each trace
    - extension_id : int : which extension is each trace in
    - aperture_id : int : which aperture matches this trace (e.g. MOS / IFU)
    - image_plane_id : int : on which image plane is this trace projected

    EXT 2 : BinTableHDU : Individual traces
    +++++++++++++++++++++++++++++++++++++++
    No special header keywords are required in this extension

    Required Table columns:

    - wavelength : float : [um] : wavelength of monochromatic aperture image
    - s : float : [arcsec] : position along aperture perpendicular to trace
    - x : float : [mm] : x position of aperture image on focal plane
    - y : float : [mm] : y position of aperture image on focal plane

    """
    _class_params = {"x_colname": "x",
                     "y_colname": "y",
                     "s_colname": "s",
                     "wave_colname": "wavelength",
                     "col_number_start": 0,
                     "center_on_wave_mid": False,
                     "dwave": 0.002,  # [um] for finding the best fit dispersion
                     "invalid_value": None,  # for dodgy trace file values
                     }

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        if "hdulist" in kwargs and isinstance(kwargs["hdulist"], fits.HDUList):
            self._file = kwargs["hdulist"]

        params = {"z_order": [70, 270, 670],
                  "pixel_scale": "!INST.pixel_scale",  # [arcsec / pix]}
                  "plate_scale": "!INST.plate_scale",  # [arcsec / mm]
                  "wave_min": "!SIM.spectral.wave_min",  # [um]
                  "wave_mid": "!SIM.spectral.wave_mid",  # [um]
                  "wave_max": "!SIM.spectral.wave_max",  # [um]
                  "x_colname": "x",
                  "y_colname": "y",
                  "s_colname": "s",
                  "wave_colname": "wavelength",
                  "center_on_wave_mid": False,
                  "dwave": 0.002,  # [um] for finding the best fit dispersion
                  "invalid_value": None,  # for dodgy trace file values
                  "report_plot_include": True,
                  "report_table_include": False,
                  }
        self.meta.update(params)

        # Parameters that are specific to the subclass
        self.meta.update(self._class_params)
        self.meta.update(kwargs)

        if self._file is not None:
            self.make_spectral_traces()

    def make_spectral_traces(self):
        """Returns a dictionary of spectral traces read in from a file"""
        self.ext_data = self._file[0].header["EDATA"]
        self.ext_cat = self._file[0].header["ECAT"]
        self.catalog = Table(self._file[self.ext_cat].data)

        spec_traces = {}
        for row in self.catalog:
            params = {col: row[col] for col in row.colnames}
            params.update(self.meta)
            hdu = self._file[row["extension_id"]]
            spec_traces[row["description"]] = SpectralTrace(hdu, **params)

        self.spectral_traces = spec_traces

    def apply_to(self, obj, **kwargs):
        """
        Interface between FieldOfView and SpectralTraceList

        This is called twice:
        1. During setup of the required FieldOfView objects, the
        SpectralTraceList is asked for the source space volumes that
        it requires (spatial limits and wavelength limits).
        2. During "observation" the method is passed a single FieldOfView
        object and applies the mapping to the image plane to it.
        The FieldOfView object is associated to one SpectralTrace from the
        list, identified by meta['trace_id'].
        """
        if isinstance(obj, FOVSetupBase):
            # Setup of FieldOfView object
            volumes = [self.spectral_traces[key].fov_grid()
                       for key in self.spectral_traces]
            new_vols_list = []
            for vol in volumes:
                wave_edges = [vol["wave_min"], vol["wave_max"]]
                if "x_min" in vol:
                    x_edges = [vol["x_min"], vol["x_max"]]
                    y_edges = [vol["y_min"], vol["y_max"]]
                    extracted_vols = obj.extract(axes=["wave", "x", "y"],
                                                 edges=(wave_edges,
                                                        x_edges,
                                                        y_edges),
                                                 aperture_id=vol["aperture_id"])
                else:
                    extracted_vols = obj.extract(axes=["wave"],
                                                 edges=(wave_edges, ),
                                                 aperture_id=vol["aperture_id"])

                for ex_vol in extracted_vols:
                    ex_vol["meta"].update(vol)
                    ex_vol["meta"].pop("wave_min")
                    ex_vol["meta"].pop("wave_max")
                new_vols_list += extracted_vols

            obj.volumes = new_vols_list

        if isinstance(obj, FieldOfViewBase):
            # Application to field of view
            if obj.hdu is not None and obj.hdu.header["NAXIS"] == 3:
                obj.cube = obj.hdu
            elif obj.hdu is not None and obj.hdu.header["NAXIS"] == 2:
                # todo: catch the case of obj.hdu.header["NAXIS"] == 2
                # for MAAT
                pass
            elif obj.hdu is None and obj.cube is None:
                obj.cube = obj.make_cube_hdu()

            # ..todo: obj will be changed to a single one covering the full field of view
            # covered by the image slicer (28 slices for LMS; for LSS still only a single slit)
            # We need a loop over spectral_traces that chops up obj into the single-slice fov before
            # calling map_spectra...
            trace_id = obj.meta['trace_id']
            spt = self.spectral_traces[trace_id]
            obj.hdu = spt.map_spectra_to_focal_plane(obj)

        return obj

    def get_waveset(self, pixel_size=None):
        if pixel_size is None:
            pixel_size = self.meta["pixel_scale"] / self.meta["plate_scale"]

        wavesets = [spt.get_pixel_wavelength_edges(pixel_size)
                    for spt in self.spectral_traces]

        return wavesets

    def get_fov_headers(self, sky_header, **kwargs):
        fov_headers = []
        for spt in self.spectral_traces:
            fov_headers += spt.fov_headers(sky_header=sky_header, **kwargs)

        return fov_headers


    @property
    def footprint(self):
        """Return the footprint of the entire SpectralTraceList"""
        xfoot, yfoot = [], []
        for spt in self.spectral_traces.values():
            xtrace, ytrace = spt.footprint()
            xfoot += xtrace
            yfoot += ytrace

        xfoot = [np.min(xfoot), np.max(xfoot), np.max(xfoot), np.min(xfoot)]
        yfoot = [np.min(yfoot), np.min(yfoot), np.max(yfoot), np.max(yfoot)]

        return xfoot, yfoot

    @property
    def image_plane_header(self):
        x, y = self.footprint
        pixel_scale = from_currsys(self.meta["pixel_scale"])
        hdr = header_from_list_of_xy(x, y, pixel_scale, "D")

        return hdr

    def plot(self, wave_min=None, wave_max=None, **kwargs):
        if wave_min is None:
            wave_min = from_currsys("!SIM.spectral.wave_min")
        if wave_max is None:
            wave_max = from_currsys("!SIM.spectral.wave_max")

        from matplotlib import pyplot as plt
        from matplotlib._pylab_helpers import Gcf
        if len(Gcf.figs) == 0:
            plt.figure(figsize=(12, 12))

        if self.spectral_traces is not None:
            clrs = "rgbcymk" * (1 + len(self.spectral_traces) // 7)
            for spt, c in zip(self.spectral_traces.values(), clrs):
                spt.plot(wave_min, wave_max, c=c)

        return plt.gcf()

    def __repr__(self):
        return "\n".join([spt.__repr__() for spt in self.spectral_traces])

    def __str__(self):
        msg = 'SpectralTraceList: "{}" : {} traces' \
              ''.format(self.meta.get("name"), len(self.spectral_traces))
        return msg


class SpectralTraceListWheel(Effect):
    """
    A Wheel-Effect object for selecting between multiple gratings/grisms

    See ``SpectralTraceList`` for the trace file format description.

    Parameters
    ----------
    trace_list_names : list
        The list of unique identifiers in the trace filenames

    filename_format : str
        ``f-string`` that directs scopesim to the folder containing the trace
        files. This can be a ``!-string`` if the trace names are shared with
        other ``*Wheel`` effect objects (e.g. a ``FilterWheel``). See examples.

    current_trace_list : str
        default trace file to use

    kwargs : key-value pairs
        Addition keywords that are passed to the ``SpectralTraceList`` objects
        See SpectralTraceList docstring

    Examples
    --------
    A simplified YAML file example taken from the OSIRIS instrument package::

        alias: INST
        name: OSIRIS_LSS

        properties:
          decouple_detector_from_sky_headers: True
          grism_names:
            - R300B
            - R500B
            - R1000B
            - R2500V

        effects:
          - name: spectral_trace_wheel
            description: grism wheel contining spectral trace geometries
            class: SpectralTraceListWheel
            kwargs:
              current_trace_list: "!OBS.grating_name"
              filename_format: "traces/LSS_{}_TRACE.fits"
              trace_list_names: "!INST.grism_names"

          - name: grating_efficiency
            description: OSIRIS grating efficiency curves, piggybacking on FilterWheel
            class: FilterWheel
            kwargs:
              minimum_throughput: !!float 0.
              filename_format: "gratings/{}.txt"
              current_filter: "!OBS.grating_name"
              filter_names: "!INST.grism_names"

    """
    def __init__(self, **kwargs):
        required_keys = ["trace_list_names",
                         "filename_format",
                         "current_trace_list"]
        check_keys(kwargs, required_keys, action="error")

        super(SpectralTraceListWheel, self).__init__(**kwargs)

        params = {"z_order": [70, 270, 670],
                  "path": "",
                  "report_plot_include": True,
                  "report_table_include": True,
                  "report_table_rounding": 4}
        self.meta.update(params)
        self.meta.update(kwargs)

        path = pth.join(self.meta["path"],
                        from_currsys(self.meta["filename_format"]))
        self.trace_lists = {}
        for name in from_currsys(self.meta["trace_list_names"]):
            kwargs["name"] = name
            self.trace_lists[name] = SpectralTraceList(filename=path.format(name),
                                                       **kwargs)

    def apply_to(self, obj, **kwargs):
        """Use apply_to of current trace list"""
        return self.current_trace_list.apply_to(obj, **kwargs)

    @property
    def current_trace_list(self):
        trace_list_eff = None
        trace_list_name = from_currsys(self.meta["current_trace_list"])
        if trace_list_name is not None:
            trace_list_eff = self.trace_lists[trace_list_name]
        return trace_list_eff

    @property
    def display_name(self):
        name = self.meta.get("name", self.meta.get("filename", "<untitled>"))
        return f'{name} : [{from_currsys(self.meta["current_trace_list"])}]'


import logging

import numpy as np
from astropy import units as u
from astropy.table import Table
from astropy.utils.data import download_file
from astropy.io import ascii as ioascii
from astropy.wcs import WCS
from synphot import SpectralElement, SourceSpectrum, Empirical1D, Observation
from synphot.units import PHOTLAM

from scopesim.source.source_templates import vega_spectrum, st_spectrum, \
    ab_spectrum
from ..utils import find_file, quantity_from_table, from_currsys

FILTER_DEFAULTS = {"U": "Generic/Bessell.U",
                   "B": "Generic/Bessell.B",
                   "V": "Generic/Bessell.V",
                   "R": "Generic/Bessell.R",
                   "I": "Generic/Bessell.I",
                   "J": "2MASS/2MASS.J",
                   "H": "2MASS/2MASS.H",
                   "Ks": "2MASS/2MASS.Ks",
                   "K": "Generic/Johnson_UBVRIJHKL.K",
                   "L": "Gemini/NIRI.Lprime-G0207w",
                   "M": "Gemini/NIRI.Mprime-G0208w",
                   "N": "Generic/Johnson_UBVRIJHKL.N",
                   "u": "SLOAN/SDSS.u",
                   "g": "SLOAN/SDSS.g",
                   "r": "SLOAN/SDSS.r",
                   "i": "SLOAN/SDSS.i",
                   "z": "SLOAN/SDSS.z",
                   "u'": "SLOAN/SDSS.uprime_filter",
                   "g'": "SLOAN/SDSS.gprime_filter",
                   "r'": "SLOAN/SDSS.rprime_filter",
                   "i'": "SLOAN/SDSS.iprime_filter",
                   "z'": "SLOAN/SDSS.zprime_filter",
                   "HAlpha": "Gemini/GMOS-N.Ha",
                   "PaBeta": "Gemini/NIRI.PaBeta-G0221",
                   "BrGamma": "Gemini/NIRI.BrG-G0218",
                   }


def get_filter_effective_wavelength(filter_name):
    if isinstance(filter_name, str):
        filter_name = from_currsys(filter_name)
        wave, trans = download_svo_filter(FILTER_DEFAULTS[filter_name],
                                          return_style="quantity")
        eff_wave = np.sum(wave * trans) / np.sum(trans)      # convert from Angstrom
        eff_wave = eff_wave.to(u.um)
    else:
        eff_wave = filter_name

    return eff_wave


def download_svo_filter(filter_name, return_style="synphot",
                        error_on_wrong_name=True):
    """
    Query the SVO service for the true transmittance for a given filter

    Copied 1 to 1 from tynt by Brett Morris

    Parameters
    ----------
    filter_name : str
        Name of the filter as available on the spanish VO filter service
        e.g: ``Paranal/HAWKI.Ks``

    return_style : str, optional
        Defines the format the data is returned
        - synphot: synphot.SpectralElement
        - table: astropy.table.Table
        - quantity: astropy.unit.Quantity [wave, trans]
        - array: np.ndarray [wave, trans], where wave is in Angstrom
        - vo_table : astropy.table.Table - original output from SVO service

    error_on_wrong_name : bool
        Default True. Raises an exception if filter_name is as incorrect SVO ID

    Returns
    -------
    filt_curve : See return_style
        Astronomical filter object.

    """
    url = f"http://svo2.cab.inta-csic.es/theory/fps3/fps.php?ID={filter_name}"
    path = download_file(url, cache=True)

    try:
        tbl = Table.read(path, format='votable')
        wave = u.Quantity(tbl['Wavelength'].data.data, u.Angstrom, copy=False)
        trans = tbl['Transmission'].data.data
    except:
        if error_on_wrong_name:
            raise ValueError(f"{filter_name} is an incorrect SVO identiier")
        else:
            logging.warning(f"'{filter_name}' was not found in the SVO. "
                            f"Defaulting to a unity transmission curve.")
            wave = [3e3, 3e5] << u.Angstrom
            trans = np.array([1., 1.])

    if return_style == "synphot":
        filt = SpectralElement(Empirical1D, points=wave, lookup_table=trans)
    elif return_style == "table":
        filt = Table(data=[wave, trans], names=["wavelength", "transmission"])
        filt.meta["wavelength_unit"] = "Angstrom"
    elif return_style == "quantity":
        filt = [wave, trans]
    elif return_style == "array":
        filt = [wave.value, trans]
    elif return_style == "vo_table":
        filt = tbl

    return filt


def download_svo_filter_list(observatory, instrument, short_names=False,
                             include=None, exclude=None):
    """
    Query the SVO service for a list of filter names for an instrument

    Parameters
    ----------
    observatory : str
        Name of the observatory as available on the spanish VO filter service
        e.g: ``Paranal/HAWKI.Ks`` --> Paranal

    instrument : str
        Name of the instrument. Be careful of hyphens etc. E.g. "HAWK-I"

    short_names : bool
        Default False. If True, the full SVO names (obs/inst.filt) are split to
        only return the (filt) part of the name

    include, exclude: str
        Each a string sequence for excluding or including specific filters
        E.g. GTC/OSIRIS has curves for ``sdss_g`` and ``sdss_g_filter``.
        We can force the inclusion of only the filter curves by setting
        ``include="_filter"``.

    Returns
    -------
    names : list
        A list of filter names

    """
    base_url = f"http://svo2.cab.inta-csic.es/theory/fps3/fps.php?"
    url = base_url + f"Facility={observatory}&Instrument={instrument}"
    path = download_file(url, cache=True)

    tbl = Table.read(path, format='votable')
    names = list(tbl["filterID"])
    if short_names:
        names = [name.split(".")[-1] for name in names]
    if include is not None:
        names = [name for name in names if include in name]
    if exclude is not None:
        names = [name for name in names if exclude not in name]

    return names


def get_filter(filter_name):
    # first check locally
    # check generics
    # check spanish_vo
    path = find_file(filter_name, silent=True)

    if path is not None:
        tbl = ioascii.read(path)
        wave = quantity_from_table("wavelength", tbl, u.um).to(u.um)
        filt = SpectralElement(Empirical1D, points=wave,
                               lookup_table=tbl["transmission"])
    elif filter_name in FILTER_DEFAULTS:
        filt = download_svo_filter(FILTER_DEFAULTS[filter_name])
    else:
        try:
            filt = download_svo_filter(filter_name)
        except:
            filt = None

    return filt


def get_zero_mag_spectrum(system_name="AB"):
    if system_name.lower() in ["vega"]:
        spec = vega_spectrum()
    elif system_name.lower() in ["ab"]:
        spec = ab_spectrum()
    elif system_name.lower() in ["st", "hst"]:
        spec = st_spectrum()

    return spec


def zero_mag_flux(filter_name, photometric_system, return_filter=False):
    """
    Returns the zero magnitude photon flux for a filter

    Acceptable filter names are those given in
    ``scopesim.effects.ter_curves_utils.FILTER_DEFAULTS`` or a string with an
    appropriate name for a filter in the Spanish-VO filter-service. Such strings
    must use the naming convention: observatory/instrument.filter. E.g:
    ``paranal/HAWKI.Ks``, or ``Gemini/GMOS-N.CaT``.

    Parameters
    ----------
    filter_name : str
        Name of the filter - see above

    photometric_system : str
        ["vega", "AB", "ST"] Name of the photometric system

    return_filter : bool, optional
        If True, also returns the filter curve object

    Returns
    -------
    flux : float
        [PHOTLAM]
    filt : ``synphot.SpectralElement``
        If ``return_filter`` is True

    """

    filt = get_filter(filter_name)
    spec = get_zero_mag_spectrum(photometric_system)

    obs = Observation(spec, filt)
    flux = obs.effstim(flux_unit=PHOTLAM)

    if return_filter:
        return flux, filt
    else:
        return flux


def scale_spectrum(spectrum, filter_name, amplitude):
    """
    Scales a SourceSpectrum to a value in a filter

    Parameters
    ----------
    spectrum : synphot.SourceSpectrum

    filter_name : str
        Name of a filter from
        - a local instrument package (available in ``rc.__search_path__``)
        - a generic filter name (see ``ter_curves_utils.FILTER_DEFAULTS``)
        - a spanish-vo filter service reference (e.g. ``"Paranal/HAWKI.Ks"``)

    amplitude : ``astropy.Quantity``, float
        The value that the spectrum should have in the given filter. Acceptable
        astropy quantities are:
        - u.mag : Vega magnitudes
        - u.ABmag : AB magnitudes
        - u.STmag : HST magnitudes
        - u.Jy : Jansky per filter bandpass
        Additionally the ``FLAM`` and ``FNU`` units from ``synphot.units`` can
        be used when passing the quantity for ``amplitude``:

    Returns
    -------
    spectrum : synphot.SourceSpectrum
        Input spectrum scaled to the given amplitude in the given filter

    Examples
    --------
    ::

        >>> from scopesim.source.source_templates import vega_spectrum
        >>> from scopesim.effects.ter_curves_utils as ter_utils
        >>>
        >>> spec = vega_spectrum()
        >>> vega_185 = ter_utils.scale_spectrum(spec, "Ks", -1.85 * u.mag)
        >>> ab_0 = ter_utils.scale_spectrum(spec, "Ks", 0 * u.ABmag)
        >>> jy_3630 = ter_utils.scale_spectrum(spec, "Ks", 3630 * u.Jy)

    """

    if isinstance(amplitude, u.Quantity):
        if amplitude.unit.physical_type == "spectral flux density":
            if amplitude.unit != u.ABmag:
                amplitude = amplitude.to(u.ABmag)
            ref_spec = ab_spectrum(amplitude.value)

        elif amplitude.unit.physical_type == "spectral flux density wav":
            if amplitude.unit != u.STmag:
                amplitude = amplitude.to(u.STmag)
            ref_spec = st_spectrum(amplitude.value)

        elif amplitude.unit == u.mag:
            ref_spec = vega_spectrum(amplitude.value)

        else:
            raise ValueError(f"Units of amplitude must be one of "
                             f"[u.mag, u.ABmag, u.STmag, u.Jy]: {amplitude}")
    else:
        ref_spec = vega_spectrum(amplitude)

    filt = get_filter(filter_name)
    ref_flux = Observation(ref_spec, filt).effstim(flux_unit=PHOTLAM)

    real_flux = Observation(spectrum, filt).effstim(flux_unit=PHOTLAM)
    scale_factor = ref_flux / real_flux
    spectrum *= scale_factor.value

    return spectrum

def apply_throughput_to_cube(cube, thru):
    """
    Apply throughput curve to a spectroscopic cube

    Parameters
    ----------
    cube : ImageHDU
         three-dimensional image, dimension 0 (in python convention) is the
         spectral dimension. WCS is required.
    thru : synphot.SpectralElement, synphot.SourceSpectrum

    Returns
    -------
    cube : ImageHDU, header unchanged, data multiplied with wavelength-dependent
         throughput
    """
    wcs = WCS(cube.header).spectral
    wave_cube = wcs.all_pix2world(np.arange(cube.data.shape[0]), 0)[0]
    wave_cube = (wave_cube * u.Unit(wcs.wcs.cunit[0])).to(u.AA)
    cube.data *= thru(wave_cube).value[:, None, None]
    return cube

def combine_two_spectra(spec_a, spec_b, action, wave_min, wave_max):
    """
    Combines transmission and/or emission spectrum with a common waverange

    Spec_A is the source spectrum
    Spec_B is either the transmission or emission that should be applied

    Parameters
    ----------
    spec_a : synphot.SourceSpectrum
    spec_b : synphot.SpectralElement, synphot.SourceSpectrum
    action: str
        ["multiply", "add"]
    wave_min, wave_max : quantity
        [Angstrom]

    Returns
    -------
    new_source : synphot.SourceSpectrum

    """
    if spec_a.waveset is None:
        wave_val = spec_b.waveset.value
    else:
        wave_val = spec_a.waveset.value
    mask = (wave_val > wave_min.value) * (wave_val < wave_max.value)

    wave = ([wave_min.value] + list(wave_val[mask]) + [wave_max.value]) * u.AA
    if "mult" in action.lower():
        spec_c = spec_a(wave) * spec_b(wave)
        ## Diagnostic plots - not for general use
        # from matplotlib import pyplot as plt
        # plt.plot(wave, spec_a(wave), label="spec_a")
        # plt.plot(wave, spec_b(wave), label="spec_b")
        # plt.plot(wave, spec_c, label="spec_c")
        # plt.xlim(2.9e4, 4.2e4)
        # plt.legend()
        # plt.show()
    elif "add" in action.lower():
        spec_c = spec_a(wave) + spec_b(wave)

    new_source = SourceSpectrum(Empirical1D, points=wave, lookup_table=spec_c)
    new_source.meta.update(spec_b.meta)
    new_source.meta.update(spec_a.meta)

    return new_source


def add_edge_zeros(tbl, wave_colname):
    if isinstance(tbl, Table):
        vals = np.zeros(len(tbl.colnames))
        col_i = np.where(col == wave_colname for col in tbl.colnames)[0][0]
        sgn = np.sign(np.diff(tbl[wave_colname][:2]))
        vals[col_i] = tbl[wave_colname][0] * (1 - 1e-7 * sgn)
        tbl.insert_row(0, vals)
        vals[col_i] = tbl[wave_colname][-1] * (1 + 1e-7 * sgn)
        tbl.insert_row(len(tbl), vals)

    return tbl



version = '0.5.4'
date    = '2022-10-06 16:00:00 GMT'
yaml_descriptions = """
- version : 0.5.4
  date : 2022-10-06
  comment : Hotfix for header keyword generators
  github_pr_url : https://github.com/AstarVienna/ScopeSim/pull/166
  changes :
  - incremental special characters for header keywords changed from `§`to `++`
  - source object function calls are now given their own FITS header keyword FNSRCn (function-call source N) due to astropy not liking the combination of HIERARCH and CONTINUE keywords

- version : 0.5.3
  date : 2022-09-29
  comment : Minor upgrade to Spec modes and to FITS keywords
  github_pr_url : https://github.com/AstarVienna/ScopeSim/pull/165
  changes :
  - Effect object ExtraFitsKeywords now has the ability to add keywords with incrementing index numbers based on the extension number
  - FOV + FOVManager + FOVVolumes classes now accept aperture_id as an argument
  - ApertureList effects object now has an apply_to function which splits the FOVVolumeList accordingly

- version : 0.5.2
  date : 2022-08-25
  comment : Update of DLC server URL to scopesim.univie.ac.at
  changes :
  - Updated MANIFEST.in to include all the files needed by the basic_instrument test optical train
  - Small update to allow iterative extension specific FITS header keywords. E.g. EXTNAME = DETn.DATA

- version : 0.5.1
  date : 2022-07-12
  comment : Update of DLC server URL to scopesim.univie.ac.at
  changes :
  - Changed URL in defaults.yaml file

- version : 0.5.0
  date : 2022-04-22
  comment : IFU Spectroscopy mode for METIS
  changes :
  - The IFU effects for the METIS LMS mode
  - Effects for including extra FITS keywords for the MICADO-ESO delivery
  - Minor change to the OpticalTrain.readout method to allow custom FITS keywords to be added
  - #-strings for accessing the .meta dict contents of Effect objects
  - added the tests.mocks.basic_instrument package for test purposes
  - refactored package downloading functions in server.database 
  - Packages can now be downloaded directly from a git commit 
  - new RTDs structure for docs based on ipynb files
  - change to SkyCalcTERCurve to use local files to avoid calling the skycalc server  
  - New Effects:
    - MetisLMSSpectralTraceList(SpectralTraceList)
    - MetisLMSSpectralTrace(SpectralTrace)
    - MetisLMSImageSlicer(ApertureMask)
    - MetisLMSEfficiency(TERCurve)
    - ExtraFitsKeywords(Effect)
    - EffectsMetaKeywords(ExtraFitsKeywords)
    - SourceDescriptionFitsKeywords(ExtraFitsKeywords)
    - SimulationConfigFitsKeywords(ExtraFitsKeywords)
    - SpectralTraceListWheel(Effect)
    - Bias(Effect)
    
- version : 0.4.1rc1
  date : 2022-03-25
  comment : Updates since METIS science team release
  changes :
  - New Effects:
    - TopHatFilterCurve   
    - TopHatFilterWheel   
    - SpanishVOFilterWheel
  - DetectorList x(y)_size columns now accept units of pixel and mm
  - warnings and errors now handled using python logging package
  - minor bug fixes

- version : 0.4.0
  date : 2022-03-03
  comment : Version released for the METIS science team
  changes :
  - release of new spectroscopy effect SpectalTraceList
  - moved individual spectral trace code to utils file
  - rewritten FovManager class
  - added make_cube, make_image, make_spectrum to FieldOfView class
  - removed fov_grid from Effects
  - added new detector array z_order section (900-class effects)
  - wavelength-dependent PSF in spectroscopic modes
  - proper handling of cube sources
  - headers for output files
  
"""



from .effects import *

from .apertures import *
from .psfs import *
from .shifts import *

from .obs_strategies import *
from .spectral_trace_list import *
from .metis_lms_trace_list import *
from .surface_list import *
from .ter_curves import *
from . import ter_curves_utils

from .detector_list import *
from .electronic import *
from .obs_strategies import *
from .fits_headers import *

# from . import effects_utils



"""
NGHXRG by Bernard Rauscher  
see the paper: http://arxiv.org/abs/1509.06264  
downloaded from: http://jwst.nasa.gov/publications.html
"""

# dependencies include: astropy, numpy, scipy [, datetime, warnings, os]
import os
import datetime
# import warnings

import numpy as np
from scipy.ndimage.interpolation import zoom
from astropy.io import fits
from astropy.stats.funcs import median_absolute_deviation as mad

# import matplotlib.pyplot as plt # Handy for debugging

#warnings.filterwarnings('ignore')

__all__ = []

class HXRGNoise:
    """
    A class to generate HxRG noise frames
    
    HXRGNoise is a class for making realistic Teledyne HxRG system
    noise. The noise model includes correlated, uncorrelated,
    stationary, and non-stationary components. The default parameters
    make noise that resembles Channel 1 of JWST NIRSpec. NIRSpec uses
    H2RG detectors. They are read out using four video outputs at
    100.000 pix/s/output.
    """

    # These class variables are common to all HxRG detectors
    nghxrg_version = 2.3 # Software version

    def __init__(self, naxis1=None, naxis2=None, naxis3=None, n_out=None,
                 dt=None, nroh=None, nfoh=None, pca0_file=None, verbose=False,
                 reverse_scan_direction=False,
                 reference_pixel_border_width=None):
        """
        Simulate Teledyne HxRG+SIDECAR ASIC system noise.

        Parameters
        ----------
        naxis1 : int
            X-dimension of the FITS cube
        naxis2 : int
            Y-dimension of the FITS cube
        naxis3 : int
            Z-dimension of the FITS cube (number of up-the-ramp samples)
        n_out : int
            Number of detector outputs
        nfoh : int
            New frame overhead in rows. This allows for a short wait at the end 
            of a frame before starting the next one.
        nroh : int
            New row overhead in pixels. This allows for a short
            wait at the end of a row before starting the next one.
        dt : int
            Pixel dwell time in seconds
        pca0_file : str
            Name of a FITS file that contains PCA-zero
        verbose : bool
            Enable this to provide status reporting
        reference_pixel_border_width : int 
            Width of reference pixel border around image area
        reverse_scan_direction : bool
            Enable this to reverse the fast scanner readout directions. This
            capability was added to support Teledyne's programmable fast scan
            readout directions. The default setting =False corresponds to
            what HxRG detectors default to upon power up.
        """

        # ======================================================================
        #
        # DEFAULT CLOCKING PARAMETERS
        #
        # The following parameters define the default HxRG clocking pattern. The
        # parameters that define the default noise model are defined in the
        # mknoise() method.
        #
        # ======================================================================

        # Default clocking pattern is JWST NIRSpec
        self.naxis1    = 2048  if naxis1   is None else int(naxis1)
        self.naxis2    = 2048  if naxis2   is None else int(naxis2)
        self.naxis3    = 1     if naxis3   is None else int(naxis3)
        self.n_out     = 4     if n_out    is None else int(n_out)
        self.dt        = 1.e-5 if dt       is None else dt
        self.nroh      = 12    if nroh     is None else int(nroh)
        self.nfoh      = 1     if nfoh     is None else int(nfoh)
        self.reference_pixel_border_width = 4 \
                                            if reference_pixel_border_width is \
                                            None else reference_pixel_border_width

        # Initialize PCA-zero file and make sure that it exists and is a file
        self.pca0_file = os.getenv('NGHXRG_HOME')+'/nirspec_pca0.fits' if \
                         pca0_file is None else pca0_file
        if os.path.isfile(self.pca0_file) is False:
            print('There was an error finding pca0_file! Check to be')
            print('sure that the NGHXRG_HOME shell environment')
            print('variable is set correctly and that the')
            print('$NGHXRG_HOME/ directory contains the desired PCA0')
            print('file. The default is nirspec_pca0.fits.')
            os.sys.exit()


        # ======================================================================

        # Configure status reporting
        self.verbose = verbose

        # Configure readout direction
        self.reverse_scan_direction = reverse_scan_direction

        # Compute the number of pixels in the fast-scan direction per
        # output
        self.xsize = self.naxis1 // self.n_out

        # Compute the number of time steps per integration, per
        # output
        self.nstep = (self.xsize+self.nroh) * (self.naxis2+self.nfoh)\
                     * self.naxis3

        # For adding in ACN, it is handy to have masks of the even
        # and odd pixels on one output neglecting any gaps
        self.m_even = np.zeros((self.naxis3,self.naxis2,self.xsize))
        self.m_odd = np.zeros_like(self.m_even)
        for x in np.arange(0,self.xsize,2):
            self.m_even[:,:self.naxis2,x] = 1
            self.m_odd[:,:self.naxis2,x+1] = 1
        self.m_even = np.reshape(self.m_even, np.size(self.m_even))
        self.m_odd = np.reshape(self.m_odd, np.size(self.m_odd))

        # Also for adding in ACN, we need a mask that point to just
        # the real pixels in ordered vectors of just the even or odd
        # pixels
        self.m_short = np.zeros((self.naxis3, self.naxis2+self.nfoh, \
                                      (self.xsize+self.nroh)//2))
        self.m_short[:,:self.naxis2,:self.xsize//2] = 1
        self.m_short = np.reshape(self.m_short, np.size(self.m_short))

        # Define frequency arrays
        self.f1 = np.fft.rfftfreq(self.nstep)   # Frequencies for nstep elements
        self.f2 = np.fft.rfftfreq(2*self.nstep)  # ... for 2*nstep elements

        # Define pinkening filters. F1 and p_filter1 are used to
        # generate ACN. F2 and p_filter2 are used to generate 1/f noise.
        self.alpha = -1 # Hard code for 1/f noise until proven otherwise
        self.p_filter1 = np.sqrt(self.f1**self.alpha)
        self.p_filter2 = np.sqrt(self.f2**self.alpha)
        self.p_filter1[0] = 0.
        self.p_filter2[0] = 0.

        # Initialize pca0. This includes scaling to the correct size,
        # zero offsetting, and renormalization. We use robust statistics
        # because pca0 is real data
        hdu = fits.open(self.pca0_file)
        naxis1 = hdu[0].header['naxis1']
        naxis2 = hdu[0].header['naxis2']
        if (naxis1 != self.naxis1 or naxis2 != self.naxis2):
            zoom_factor = self.naxis1 / naxis1
            self.pca0 = zoom(hdu[0].data, zoom_factor, order=1, mode='wrap')
        else:
            self.pca0 = hdu[0].data
        self.pca0 -= np.median(self.pca0) # Zero offset
        self.pca0 /= (1.4826*mad(self.pca0)) # Renormalize

    def message(self, message_text):
        """
        Used for status reporting
        """
        if self.verbose is True:
            print('NG: ' + message_text + ' at DATETIME = ', \
                  datetime.datetime.now().time())

    def white_noise(self, nstep=None):
        """
        Generate white noise for an HxRG including all time steps
        (actual pixels and overheads).

        Parameters
        ----------
        nstep : int
            Length of vector returned
        """
        return(np.random.standard_normal(nstep))

    def pink_noise(self, mode):
        """
        Generate a vector of non-periodic pink noise.

        Parameters
        ----------
        mode : str
            Selected from {'pink', 'acn'}
        """

        # Configure depending on mode setting
        if mode is 'pink':
            nstep = 2*self.nstep
            f = self.f2
            p_filter = self.p_filter2
        else:
            nstep = self.nstep
            f = self.f1
            p_filter = self.p_filter1

        # Generate seed noise
        mynoise = self.white_noise(nstep)

        # Save the mean and standard deviation of the first
        # half. These are restored later. We do not subtract the mean
        # here. This happens when we multiply the FFT by the pinkening
        # filter which has no power at f=0.
        the_mean = np.mean(mynoise[:nstep//2])
        the_std = np.std(mynoise[:nstep//2])

        # Apply the pinkening filter.
        thefft = np.fft.rfft(mynoise)
        thefft = np.multiply(thefft, p_filter)
        result = np.fft.irfft(thefft)
        result = result[:nstep//2] # Keep 1st half

        # Restore the mean and standard deviation
        result *= the_std / np.std(result)
        result = result - np.mean(result) + the_mean

        # Done
        return(result)

    def mknoise(self, o_file, rd_noise=None, pedestal=None, c_pink=None,
                u_pink=None, acn=None, pca0_amp=None,
                reference_pixel_noise_ratio=None, ktc_noise=None,
                bias_offset=None, bias_amp=None):
        """
        Generate a FITS cube containing only noise.

        Parameters
        ----------
        o_file : str
            Output filename
        pedestal : float
            Magnitude of pedestal drift in electrons
        rd_noise : float
            Standard deviation of read noise in electrons
        c_pink : float
            Standard deviation of correlated pink noise in electrons
        u_pink : float
            Standard deviation of uncorrelated pink noise in electrons
        acn: float
            Standard deviation of alterating column noise in electrons
        pca0 : float
            Standard deviation of pca0 in electrons
        reference_pixel_noise_ratio : float
            Ratio of the standard deviation of
            the reference pixels to the regular
            pixels. Reference pixels are usually
            a little lower noise.
        ktc_noise : float
            kTC noise in electrons. Set this equal to
            sqrt(k*T*C_pixel)/q_e, where k is Boltzmann's
            constant, T is detector temperature, and C_pixel is
            pixel capacitance. For an H2RG, the pixel capacitance
            is typically about 40 fF.
        bias_offset : float
            On average, integrations stare here in electrons. Set this so that 
            all pixels are in range.
        bias_amp : float
            A multiplicative factor that we multiply PCA-zero by
            to simulate a bias pattern. This is completely
            independent from adding in "picture frame" noise.

        Notes
        -----
        Because of the noise correlations, there is no simple way to
        predict the noise of the simulated images. However, to a
        crude first approximation, these components add in
        quadrature.

        The units in the above are mostly "electrons". This follows convention
        in the astronomical community. From a physics perspective, holes are
        actually the physical entity that is collected in Teledyne's p-on-n
        (p-type implants in n-type bulk) HgCdTe architecture.

        """

        self.message('Starting mknoise()')

        # ======================================================================
        #
        # DEFAULT NOISE PARAMETERS
        #
        # These defaults create noise similar to that seen in the JWST NIRSpec.
        #
        # ======================================================================

        self.rd_noise  = 5.2      if rd_noise     is None else rd_noise
        self.pedestal  = 4        if pedestal     is None else pedestal
        self.c_pink    = 3        if c_pink       is None else c_pink
        self.u_pink    = 1        if u_pink       is None else u_pink
        self.acn       = .5       if acn          is None else acn
        self.pca0_amp  = .2       if pca0_amp     is None else pca0_amp

        # Change this only if you know that your detector is different from a
        # typical H2RG.
        self.reference_pixel_noise_ratio = 0.8 if \
            reference_pixel_noise_ratio is None else reference_pixel_noise_ratio

        # These are used only when generating cubes. They are
        # completely removed when the data are calibrated to
        # correlated double sampling or slope images. We include
        # them in here to make more realistic looking raw cubes.
        self.ktc_noise   = 29.   if ktc_noise   is None else ktc_noise
        self.bias_offset = 5000. if bias_offset is None else bias_offset
        self.bias_amp    = 500.  if bias_amp    is None else bias_amp

        # ======================================================================

        # Initialize the result cube. For up-the-ramp integrations,
        # we also add a bias pattern. Otherwise, we assume
        # that the aim was to simulate a two dimensional correlated
        # double sampling image or slope image.
        self.message('Initializing results cube')
        result = np.zeros((self.naxis3, self.naxis2, self.naxis1), \
                          dtype=np.float32)
        if self.naxis3 > 1:
            # Inject a bias pattern and kTC noise. If there are no reference pixels,
            # we know that we are dealing with a subarray. In this case, we do not
            # inject any bias pattern for now.
            if self.reference_pixel_border_width > 0:
                bias_pattern = self.pca0*self.bias_amp + self.bias_offset
            else:
                bias_pattern = self.bias_offset

            # Add in some kTC noise. Since this should always come out
            # in calibration, we do not attempt to model it in detail.
            bias_pattern += \
                         self.ktc_noise * \
                         np.random.standard_normal((self.naxis2, self.naxis1))

            # Ensure that there are no negative pixel values. Data cubes
            # are converted to unsigned integer before writing.
            bias_pattern = np.where(bias_pattern < 0, 0, bias_pattern)

            # Add in the bias pattern
            for z in np.arange(self.naxis3):
                result[z,:,:] += bias_pattern

        # Make white read noise. This is the same for all pixels.
        self.message('Generating rd_noise')
        w = self.reference_pixel_border_width # Easier to work with
        r = self.reference_pixel_noise_ratio  # Easier to work with
        for z in np.arange(self.naxis3):
            here = np.zeros((self.naxis2, self.naxis1))
            if w > 0: # Ref. pixel border exists
                # Add both reference and regular pixels
                here[:w, :] = r * self.rd_noise * \
                              np.random.standard_normal((w, self.naxis1))
                here[-w:, :] = r * self.rd_noise * \
                               np.random.standard_normal((w, self.naxis1))
                here[:, :w] = r * self.rd_noise * \
                              np.random.standard_normal((self.naxis1, w))
                here[:, -w:] = r * self.rd_noise * \
                               np.random.standard_normal((self.naxis1, w))
                # Make noisy regular pixels
                here[w:-w, w:-w] = self.rd_noise * \
                                  np.random.standard_normal((self.naxis2-2*w,
                                                             self.naxis1-2*w))
            else: # Ref. pixel border does not exist
                # Add only regular pixels
                here = self.rd_noise * np.random.standard_normal((self.naxis2,
                                                                  self.naxis1))
            # Add the noise in to the result
            result[z, :, :] += here

        # Add correlated pink noise.
        self.message('Adding c_pink noise')
        tt = self.c_pink * self.pink_noise('pink') # tt is a temp. variable
        tt = np.reshape(tt, (self.naxis3, self.naxis2+self.nfoh, \
                             self.xsize+self.nroh))[:,:self.naxis2,:self.xsize]
        for op in np.arange(self.n_out):
            x0 = op * self.xsize
            x1 = x0 + self.xsize
            if self.reverse_scan_direction is False:
                # Teledyne's default fast-scan directions
                if np.mod(op, 2) == 0:
                    result[:, :, x0:x1] += tt
                else:
                    result[:, :, x0:x1] += tt[:, :, ::-1]
            else:
                # Reverse the fast-scan directions.
                if np.mod(op,2)==1:
                    result[:,:,x0:x1] += tt
                else:
                    result[:,:,x0:x1] += tt[:,:,::-1]

        # Add uncorrelated pink noise. Because this pink noise is stationary and
        # different for each output, we don't need to flip it.
        self.message('Adding u_pink noise')
        for op in np.arange(self.n_out):
            x0 = op * self.xsize
            x1 = x0 + self.xsize
            tt = self.u_pink * self.pink_noise('pink')
            tt = np.reshape(tt, (self.naxis3, self.naxis2+self.nfoh, \
                            self.xsize+self.nroh))[:,:self.naxis2,:self.xsize]
            result[:, :, x0:x1] += tt

        # Add ACN
        self.message('Adding acn noise')
        for op in np.arange(self.n_out):

            # Generate new pink noise for each even and odd vector.
            # We give these the abstract names 'a' and 'b' so that we
            # can use a previously worked out formula to turn them
            # back into an image section.
            a = self.acn * self.pink_noise('acn')
            b = self.acn * self.pink_noise('acn')

            # Pick out just the real pixels (i.e. ignore the gaps)
            a = a[np.where(self.m_short == 1)]
            b = b[np.where(self.m_short == 1)]

            # Reformat into an image section. This uses the formula
            # mentioned above.
            acn_cube = np.reshape(np.transpose(np.vstack((a,b))),
                                  (self.naxis3,self.naxis2,self.xsize))

            # Add in the ACN. Because pink noise is stationary, we can
            # ignore the readout directions. There is no need to flip
            # acn_cube before adding it in.
            x0 = op * self.xsize
            x1 = x0 + self.xsize
            result[:,:,x0:x1] += acn_cube

        # Add PCA-zero. The PCA-zero template is modulated by 1/f.
        if self.pca0_amp > 0:
            self.message('Adding PCA-zero "picture frame" noise')
            gamma = self.pink_noise(mode='pink')
            zoom_factor = self.naxis2 * self.naxis3 / np.size(gamma)
            gamma = zoom(gamma, zoom_factor, order=1, mode='mirror')
            gamma = np.reshape(gamma, (self.naxis3, self.naxis2))
            for z in np.arange(self.naxis3):
                for y in np.arange(self.naxis2):
                    result[z, y, :] += self.pca0_amp*self.pca0[y, :]*gamma[z, y]

        # If the data cube has only 1 frame, reformat into a 2-dimensional
        # image.
        if self.naxis3 == 1:
            self.message('Reformatting cube into image')
            result = result[0, :, :]

        # If the data cube has more than one frame, convert to unsigned
        # integer
        if self.naxis3 > 1:
            self.message('Converting to 16-bit unsigned integer')
            result = result.astype('uint16')

        # Write the result to a FITS file
        self.message('Writing FITS file')
        hdu = fits.PrimaryHDU(result)
        hdu.header.append()
        hdu.header.append(('RD_NOISE', self.rd_noise, 'Read noise'))
        hdu.header.append(('PEDESTAL', self.pedestal, 'Pedestal drifts'))
        hdu.header.append(('C_PINK', self.c_pink, 'Correlated pink'))
        hdu.header.append(('U_PINK', self.u_pink, 'Uncorrelated pink'))
        hdu.header.append(('ACN', self.acn, 'Alternating column noise'))
        hdu.header.append(('PCA0', self.pca0_amp, \
                           'PCA zero, AKA picture frame'))
        #hdu.header['HISTORY'] = 'Created_by_NGHXRG_version_' \
        #                        + str(self.nghxrg_version)

        self.message('Exiting mknoise()')

        if o_file is not None:
            hdu.writeto(o_file, clobber='True')
        return result



import inspect
from copy import deepcopy, copy
from astropy.table import Table

from .. import effects as efs


def combine_surface_effects_OLD(surface_effects):
    surflist_list = [eff for eff in surface_effects
                     if isinstance(eff, efs.SurfaceList)]
    surf_list = [eff for eff in surface_effects
                 if isinstance(eff, efs.TERCurve)]

    if len(surflist_list) == 0:
        surf = empty_surface_list()
        surf.meta["name"] = "Radiometry Table"
        surflist_list += [surf]

    new_surflist = deepcopy(surflist_list[0])
    for surflist in surflist_list[1:]:
        new_surflist.add_surface_list(surflist)

    for surf in surf_list:
        position = surf.meta["position"] if "position" in surf.meta else -1
        new_surflist.add_surface(surf, surf.meta["name"], position=position)

    new_surflist.table = new_surflist.radiometry_table.table

    return new_surflist


def combine_surface_effects(surface_effects):
    surflist_list = [eff for eff in surface_effects
                     if isinstance(eff, efs.SurfaceList)]
    surf_list = [eff for eff in surface_effects
                 if isinstance(eff, (efs.TERCurve, efs.FilterWheel))
                 and not isinstance(eff, efs.SurfaceList)]

    if len(surflist_list) == 0:
        surflist_list = [empty_surface_list(name="combined_surface_list")]

    new_surflist = copy(surflist_list[0])
    for surflist in surflist_list[1:]:
        new_surflist.add_surface_list(surflist)

    # ..todo:: should read position from the list positions in surface_effects
    for surf in surf_list:
        position = surf.meta.get("position", -1)
        new_surflist.add_surface(surf, surf.meta["name"], position=position)

    return new_surflist


def get_all_effects(effects, effect_class):
    if isinstance(effect_class, (list, tuple)):
        my_effects = []
        for eff_cls in effect_class:
            my_effects += get_all_effects(effects, eff_cls)
    else:
        my_effects = [eff for eff in effects
                      if isinstance(eff, effect_class) and eff.include]

    return my_effects


def make_effect(effect_dict, **properties):
    effect_meta_dict = {key : effect_dict[key] for key in effect_dict
                        if key not in ["class", "kwargs"]}
    effect_class_name = effect_dict["class"]
    effect_cls = getattr(efs, effect_class_name)
    # ..todo: add looking for custom effect class names from 3rd party packages

    effect_kwargs = {}
    effect_kwargs.update(effect_meta_dict)         # effect name and description
    effect_kwargs.update(properties)              # optical_element properties
    if "kwargs" in effect_dict:
        effect_kwargs.update(effect_dict["kwargs"])  # individual effect kwargs

    effect = effect_cls(**effect_kwargs)
    # effect.meta.update(effect_meta_dict)  # is this needed? Seems redundant

    return effect


def is_spectroscope(effects):
    spec_classes = (efs.SpectralTraceList,
                    efs.SpectralTraceListWheel)
    return any([isinstance(eff, spec_classes) for eff in effects])


def empty_surface_list(**kwargs):
    tbl = Table(names=["name", "outer", "inner", "angle",
                       "temperature", "action", "filename"],
                data=[["test"], [0.], [0.], [0.], [0.], ["none"], ["none"]],
                meta={"outer_unit": "m", "inner_unit": "m",
                      "angle_unit": "deg", "temperature_unit": "deg_C"})
    return efs.SurfaceList(table=tbl[:0], **kwargs)


def scopesim_effect_classes(base_effect=efs.Effect):
    members = inspect.getmembers(efs)
    efs_dict = {".".join([cls.__module__, cls.__name__]).replace("scopesim.effects.", ""): cls
                for name, cls in members
                if hasattr(cls, "__mro__") and base_effect in cls.__mro__}
    sorted_effects = {key: efs_dict[key] for key in sorted(efs_dict)}

    return sorted_effects
